import itertools
from types import MappingProxyType

from matplotlib import pyplot as plt
from matplotlib.colors import LinearSegmentedColormap
from matplotlib.collections import PatchCollection
from matplotlib.patches import Rectangle

import numpy as np
import pandas as pd
import seaborn as sns
import toolz

from genopandas.plotting.clustermap import color_annotation

DEFAULT_ALTERATIONS = (
    ('cna', [('amp', '#fd0000'),
             ('homdel', '#0005fe'),
             ('gain', '#feb6c1'),
             ('hetloss', '#90d8d8')]),
    ('mutation', [('trunc', '#000000'),
                  ('missense', '#0b8008')]),
    ('expression', [('increased', '#fe9999'),
                    ('decreased', '#6799cc')])
)  # yapf: disable

DEFAULT_STYLES = MappingProxyType({
    'cna': 'fill',
    'mutation': 'inset',
    'expression': 'border'
})


class OncoPlotter(object):
    """Plotter object that facilitates the drawing of oncoplots."""

    def __init__(self,
                 alterations=DEFAULT_ALTERATIONS,
                 alteration_styles=DEFAULT_STYLES,
                 inset_width=0.6,
                 inset_height=0.6,
                 border_width=2,
                 keep_unaltered=False):

        self._alterations = alterations
        self._alteration_styles = alteration_styles

        self._inset_height = inset_height
        self._inset_width = inset_width

        self._border_lw = border_width

        self._keep_unaltered = keep_unaltered

    def plot(self,
             data,
             sample_order=None,
             gene_order=None,
             sort=True,
             limit=10,
             ax=None,
             linewidths=1):
        """Draws oncoplot."""

        # Determine gene/sample order.
        gene_order, sample_order = self._sort_orders(
            data,
            self._alterations,
            sort=sort,
            gene_order=gene_order,
            sample_order=sample_order,
            limit=limit,
            keep_unaltered=self._keep_unaltered)

        data = data.loc[data['gene'].isin(gene_order)]

        if ax is None:
            _, ax = plt.subplots()

        # Plot background.
        self._plot_background(
            gene_order=gene_order,
            sample_order=sample_order,
            ax=ax,
            linewidths=linewidths)

        # Plot elements.
        self._plot_fill(
            data,
            gene_order=gene_order,
            sample_order=sample_order,
            ax=ax,
            linewidths=linewidths)

        self._plot_inset(
            data, gene_order=gene_order, sample_order=sample_order, ax=ax)

        self._plot_border(
            data,
            gene_order=gene_order,
            sample_order=sample_order,
            ax=ax,
            linewidths=linewidths)

        # Finish styling axes.
        self._style_axis(ax)

        return OncoPlot(
            gene_order=gene_order, sample_order=sample_order, ax=ax)

    @classmethod
    def _sort_orders(cls,
                     data,
                     alterations,
                     sort=True,
                     gene_order=None,
                     sample_order=None,
                     limit=10,
                     keep_unaltered=True):
        """Determines sort orders with given inputs."""

        if sort and (gene_order is None or sample_order is None):
            sorted_genes, sorted_samples = sort_by_alterations(
                data, alterations, gene_order=gene_order, limit=limit)

            gene_order = gene_order or sorted_genes
            sample_order = sample_order or sorted_samples
        else:
            if gene_order is None:
                gene_order = cls._get_series_order(data['gene'])[:limit]

            if sample_order is None:
                sample_order = cls._get_series_order(data['sample'])

        if not keep_unaltered:
            altered = set(data.loc[data['gene'].isin(gene_order), 'sample'])
            sample_order = [
                sample for sample in sample_order if sample in altered
            ]

        return gene_order, sample_order

    @staticmethod
    def _get_series_order(series):
        if hasattr(series, 'cat'):
            return series.cat.categories
        else:
            return list(series.unique())

    def _plot_background(self, gene_order, sample_order, ax, linewidths):
        """Plots oncoplot background."""

        null_data = pd.DataFrame(0, index=gene_order, columns=sample_order)

        sns.heatmap(
            null_data,
            cmap=discrete_cmap(['gainsboro']),
            ax=ax,
            cbar=False,
            linewidths=linewidths)

    def _plot_fill(self, data, gene_order, sample_order, ax, linewidths):
        """Plots oncoplot fill elements."""

        # Fetch alterations, subset data and map to numeric matrix.
        alterations = self._subset_for_style(style='fill')

        alteration_names = {alt_name for alt_name, _ in alterations}
        data = data.loc[data['type'].isin(alteration_names)]

        matrix, colors = self._map_to_order(data, alterations, as_matrix=True)
        matrix = matrix.reindex(index=gene_order, columns=sample_order)

        # Setup matplotlib cmap (order --> colors).
        cmap = discrete_cmap(colors)

        # Plot heatmap.
        sns.heatmap(
            matrix,
            cmap=cmap,
            ax=ax,
            vmin=0,
            vmax=len(colors),
            cbar=False,
            linewidths=linewidths)

    def _subset_for_style(self, style):
        """Subsets alterations for given style."""

        return [(alt, alt_values) for alt, alt_values in self._alterations
                if self._alteration_styles[alt] == style]

    def _map_to_order(self, data, alterations, as_matrix=False):
        """Maps data to numeric order values (keeping only highest
           priority values)."""

        order_mapping, order_colors = self._order_mapping(alterations)

        value_keys = data['type'] + '_' + data['value']
        values = value_keys.map(order_mapping)

        data_num = pd.DataFrame({
            'sample': data['sample'],
            'gene': data['gene'],
            'value': values
        })

        if as_matrix:
            data_num = pd.pivot_table(
                data_num,
                index='gene',
                columns='sample',
                values='value',
                aggfunc='min')
        else:
            data_num = data_num.groupby(
                ['gene', 'sample'])['value'].min().reset_index()

        return data_num, order_colors

    @staticmethod
    def _order_mapping(alterations):
        """Creates mapping of alterations to consecutive numeric values.

        Format: {type}_{value} --> {index}. Used when pivoting matrix, to ensure
        priority is followed when converting from long format to matrix.
        """

        order_mapping, colors = {}, []

        for data_type, val_order in alterations:
            for val, color in val_order:
                key = data_type + '_' + val
                order_mapping[key] = len(order_mapping)

                colors.append(color)

        return order_mapping, colors

    def _plot_inset(self, data, gene_order, sample_order, ax):
        """Plots oncoplot inset elements."""

        margin_x = (1 - self._inset_width) / 2
        margin_y = (1 - self._inset_height) / 2

        def _draw_func(grp, gene_lookup, sample_lookup, color, ax):
            patches = (Rectangle(
                xy=(sample_lookup[row.sample] + margin_x,
                    gene_lookup[row.gene] + margin_y),
                width=self._inset_width,
                height=self._inset_height) for row in grp.itertuples())

            patches = PatchCollection(patches, facecolor=color)
            ax.add_collection(patches)

        self._draw_with_lookup(
            data,
            style='inset',
            gene_order=gene_order,
            sample_order=sample_order,
            draw_func=_draw_func,
            ax=ax)

    def _draw_with_lookup(self, data, style, gene_order, sample_order,
                          draw_func, ax):
        """Helper function that draws elements using row/column lookups."""

        # Fetch alterations, subset data and map to numeric values.
        alterations = self._subset_for_style(style=style)

        alteration_names = {alt_name for alt_name, _ in alterations}
        data = data.loc[data['type'].isin(alteration_names)]

        data_ordered, colors = self._map_to_order(
            data, alterations, as_matrix=False)

        # Draw insets.
        sample_lookup = dict(zip(sample_order, range(len(sample_order))))
        gene_lookup = dict(zip(gene_order, range(len(gene_order))))

        for value, grp in data_ordered.groupby('value'):
            if len(grp) > 0:
                draw_func(
                    grp,
                    sample_lookup=sample_lookup,
                    gene_lookup=gene_lookup,
                    color=colors[int(value)],
                    ax=ax)

    def _plot_border(self, data, gene_order, sample_order, ax, linewidths):
        """Plots 'bordered' elements."""

        offset_x, offset_y = self._border_offsets(
            self._border_lw, linewidths, ax=ax)

        def _draw_func(grp, sample_lookup, gene_lookup, color, ax):
            for row in grp.itertuples():
                patch = Rectangle(
                    xy=(sample_lookup[row.sample] + offset_x,
                        gene_lookup[row.gene] + offset_y),
                    width=1. - (2 * offset_x),
                    height=1. - (2 * offset_y),
                    edgecolor=color,
                    facecolor='none',
                    lw=self._border_lw,
                    joinstyle='miter')

                ax.add_patch(patch)

        self._draw_with_lookup(
            data,
            style='border',
            gene_order=gene_order,
            sample_order=sample_order,
            draw_func=_draw_func,
            ax=ax)

    @staticmethod
    def _border_offsets(border_lw, linewidths, ax):
        # Force fig draw.
        fig = ax.figure
        fig.canvas.draw()

        # Get transform/DPI.
        transform = ax.transData.inverted().transform
        ppd = 72. / fig.dpi

        # Transform line widths to data coords.
        border_offset = (linewidths + 0.5 * border_lw) - 0.5

        transformed = ((transform((border_offset, border_offset)) -
                        transform((0, 0))) * ppd)  # yapf: disable

        offset_x, offset_y = np.abs(transformed)

        return offset_x, offset_y

    @staticmethod
    def _style_axis(ax):
        """Styles axis."""

        ax.set_xlabel('Samples')
        ax.set_ylabel('Genes')

        plt.setp(ax.get_yticklabels(), rotation=0)


class OncoPlot(object):
    """Oncoplot figure object."""

    def __init__(self, gene_order, sample_order, ax):
        self.gene_order = gene_order
        self.sample_order = sample_order
        self.ax = ax

    @property
    def fig(self):
        """Figure on which oncoplot was drawn."""
        return self.ax.figure

    def savefig(self, *args, **kwargs):
        """Saves oncoplot to file using ``figure.savefig``."""
        return self.fig.savefig(*args, **kwargs)


def oncoplot(data,
             alterations=DEFAULT_ALTERATIONS,
             alteration_styles=DEFAULT_STYLES,
             sample_order=None,
             gene_order=None,
             sort=True,
             limit=10,
             ax=None,
             linewidths=1,
             inset_width=0.6,
             inset_height=0.6,
             border_width=2,
             keep_unaltered=False):
    """Plots oncoplot for given data."""

    plotter = OncoPlotter(
        alterations=alterations,
        alteration_styles=alteration_styles,
        inset_width=inset_width,
        inset_height=inset_height,
        border_width=border_width,
        keep_unaltered=keep_unaltered)

    return plotter.plot(
        data=data,
        sample_order=sample_order,
        gene_order=gene_order,
        sort=sort,
        limit=limit,
        ax=ax,
        linewidths=linewidths)


def sort_by_alterations(data, alterations, gene_order=None, limit=10):
    """Returns sorted gene/sample orders for given data."""

    # Convert data to ordered (numeric) format for sorting.
    sort_mapping = _sort_mapping(alterations)
    value_keys = data['type'] + '_' + data['value']

    sort_data = pd.DataFrame({
        'sample': data['sample'],
        'gene': data['gene'],
        'type': data['type'],
        'value': value_keys.map(sort_mapping)
    })

    # Determine gene order.
    if gene_order is None:
        gene_ranking = (sort_data.dropna()
                        .groupby('gene')['sample'].nunique()
                        .sort_values(ascending=False))  # yapf: disable
        gene_order = list(gene_ranking.index)

    gene_order = gene_order[:limit]

    # Take most important effect per type (max).
    sort_data = sort_data.groupby(['sample', 'gene', 'type'])['value'].max()
    sort_data = sort_data.reset_index()

    # Sort, taking gene order into account.
    sort_data = sort_data.loc[sort_data['gene'].isin(gene_order)]
    sort_mat = pd.pivot_table(
        sort_data,
        index='sample',
        columns='gene',
        values='value',
        aggfunc='sum')

    sort_mat = sort_mat.sort_values(by=gene_order, ascending=False)
    sample_order = list(sort_mat.index)

    return gene_order, sample_order


def _sort_mapping(alterations):
    """Maps data types and their values to numeric values used for sorting."""

    n_levels = len(alterations)

    mapping = {}
    for i, (data_type, val_order) in enumerate(alterations):
        multiplier = 10**(n_levels - i)

        for j, (val, _) in enumerate(val_order):
            key = data_type + '_' + val
            mapping[key] = multiplier * (len(val_order) - j)

    return mapping


def discrete_cmap(colors):
    """Create an N-bin discrete colormap from specified colors."""

    if not colors:
        return None
    else:
        if len(colors) == 1:
            # Repeat single colors as a work-around
            # for single color issue.
            colors = colors * 2

        n_colors = len(colors)
        base = plt.cm.get_cmap()
        cmap_name = base.name + str(n_colors)
        return LinearSegmentedColormap.from_list(cmap_name, colors, n_colors)


class OncoprintPlotter(object):
    """Plotter object that facilitates the drawing of oncoprints."""

    def __init__(self,
                 alterations=DEFAULT_ALTERATIONS,
                 alteration_styles=DEFAULT_STYLES,
                 inset_width=0.6,
                 inset_height=0.6,
                 border_width=2,
                 keep_unaltered=False,
                 bar_color='dimgray'):

        self._alterations = alterations
        self._bar_color = bar_color

        self._oncoplotter = OncoPlotter(
            alterations=alterations,
            alteration_styles=alteration_styles,
            inset_width=inset_width,
            inset_height=inset_height,
            border_width=border_width,
            keep_unaltered=keep_unaltered)

    def plot(self,
             data,
             sample_order=None,
             gene_order=None,
             annotation=None,
             annotation_colors=None,
             sort=True,
             limit=10,
             linewidths=1,
             figsize=None,
             gridspec_kws=None):
        """Draws oncoprint on figure."""

        # Setup figure.
        fig, axes = self._setup_figure(
            figsize=figsize,
            gridspec_kws=gridspec_kws,
            with_annotation=annotation is not None)

        # Draw oncoplot.
        oncoplot = self._oncoplotter.plot(
            data=data,
            sample_order=sample_order,
            gene_order=gene_order,
            sort=sort,
            limit=limit,
            ax=axes[-1, 0],
            linewidths=linewidths)

        # Draw frequencies + annotation (if given).
        self._plot_frequencies(
            data,
            gene_order=oncoplot.gene_order,
            ax=axes[-1, 1],
            color=self._bar_color)

        if annotation is not None:
            annotation = annotation.reindex(index=oncoplot.sample_order)

            self._plot_annotation(
                annotation,
                annotation_colors,
                ax=axes[0, 0],
                linewidths=linewidths)

        # Style axes.
        axes[0, 1].axis('off')

        return Oncoprint(fig=fig, axes=axes, oncoplot=oncoplot)

    @staticmethod
    def _setup_figure(figsize, gridspec_kws=None, with_annotation=False):
        """Sets up figure + axes for plotting."""

        gridspec_defaults = {
            'width_ratios': [0.8, 0.2],
            'wspace': 0.02,
            'hspace': 0.02
        }

        if with_annotation:
            nrows = 2
            gridspec_defaults['height_ratios'] = [0.2, 0.8]
        else:
            nrows = 1

        gridspec_kws = toolz.merge(gridspec_defaults, (gridspec_kws or {}))

        fig, axes = plt.subplots(
            figsize=figsize,
            ncols=2,
            nrows=nrows,
            gridspec_kw=gridspec_kws,
            squeeze=False)

        return fig, axes

    def _plot_frequencies(self, data, gene_order, ax=None, **kwargs):
        """Plots bargraph with alteration frequencies."""

        # Subset data for alterations.
        alt_values = set(alt + '_' + value
                         for alt, alt_values in self._alterations
                         for value, _ in alt_values)

        mask = (data['type'] + '_' + data['value']).isin(alt_values)
        data = data.loc[mask]

        # Calculate frequencies/fractions.
        freqs = (data.groupby('gene')['sample'].nunique()
                 .sort_values(ascending=False))  # yapf: disable

        fractions = freqs / data['sample'].nunique()

        # Use gene_order to select shown genes.
        fractions = fractions.loc[gene_order]

        # Plot barplot.
        ax = sns.barplot(
            data=fractions.to_frame('fraction').reset_index(),
            y='gene',
            x='fraction',
            ax=ax,
            order=fractions.index,
            **kwargs)

        # Annotate with values.
        for i, value in enumerate(fractions):
            ax.text(
                value + 0.02,
                i,
                '{0:.0f}%'.format(value * 100),
                ha='left',
                va='center')

        # Style axes.
        ax.get_xaxis().set_visible(False)

        ax.set_yticks([])
        ax.set_ylabel('')

        sns.despine(ax=ax, left=True, bottom=True)

        return ax

    def _plot_annotation(self, annotation, annotation_colors, ax, linewidths):
        """Plots annotation along top, in similar style as sns.clustermap."""

        colored, _ = color_annotation(annotation, annotation_colors)
        colored_numeric, cmap = self._color_matrix_to_numeric(colored)

        sns.heatmap(
            colored_numeric.T,
            cmap=cmap,
            ax=ax,
            cbar=False,
            linewidths=linewidths,
            yticklabels=True)

        ax.set_xticks([])
        ax.set_xlabel('')

        plt.setp(ax.get_yticklabels(), rotation=0)

    @staticmethod
    def _color_matrix_to_numeric(matrix):
        """Converts matrix containing color values to a numeric matrix + cmap,
           which can be used to draw the colors using sns.heatmap."""

        # Get list of all colors in matrix.
        col_colors = [col_values.unique()
                      for _, col_values in matrix.items()]  # yapf: disable
        all_colors = list(set(itertools.chain.from_iterable(col_colors)))

        # Map colors to integer values.
        value_map = dict(zip(all_colors, range(len(all_colors))))
        num_matrix = matrix.applymap(lambda v: value_map.get(v))

        # Create cmap.
        cmap = discrete_cmap(all_colors)

        return num_matrix, cmap


def oncoprint(data,
              alterations=DEFAULT_ALTERATIONS,
              alteration_styles=DEFAULT_STYLES,
              sample_order=None,
              gene_order=None,
              annotation=None,
              annotation_colors=None,
              sort=True,
              limit=10,
              linewidths=1,
              inset_width=0.6,
              inset_height=0.6,
              border_width=2,
              keep_unaltered=False,
              bar_color='dimgray',
              figsize=None,
              gridspec_kws=None):
    """Plots oncoprint (with freq. and annotation) for given data."""

    plotter = OncoprintPlotter(
        alterations=alterations,
        alteration_styles=alteration_styles,
        inset_width=inset_width,
        inset_height=inset_height,
        border_width=border_width,
        bar_color=bar_color,
        keep_unaltered=keep_unaltered)

    return plotter.plot(
        data=data,
        annotation=annotation,
        annotation_colors=annotation_colors,
        sample_order=sample_order,
        gene_order=gene_order,
        sort=sort,
        limit=limit,
        linewidths=linewidths,
        figsize=figsize,
        gridspec_kws=gridspec_kws)


class Oncoprint(object):
    """Oncoprint figure object."""

    def __init__(self, fig, axes, oncoplot):
        self.fig = fig
        self.axes = axes
        self.oncoplot = oncoplot

    @property
    def ax_oncoplot(self):
        """Axis on which oncoplot was drawn."""
        return self.oncoplot.ax

    @property
    def ax_frequency(self):
        """Axis on which alteration frequencies were drawn."""
        return self.axes[1, 1]

    def savefig(self, *args, **kwargs):
        """Saves figure using ``fig.savefig``."""
        return self.fig.savefig(*args, **kwargs)



import numpy as np

from genopandas.core.matrix import AnnotatedMatrix


class ExpressionMatrix(AnnotatedMatrix):
    """Matrix containing (gene) expression values (features-by-samples)."""

    @classmethod
    def from_subread(cls,
                     file_path,
                     sample_data=None,
                     sample_mapping=None,
                     **kwargs):
        """Reads expression from a subread output file."""

        return super().from_csv(
            file_path,
            sample_data=sample_data,
            sample_mapping=sample_mapping,
            drop_cols=['Chr', 'Start', 'End', 'Strand', 'Length'],
            index_col=0,
            sep='\t',
            **kwargs)

    def normalize(self, size_factors=None, log2=False):
        """Normalizes expression counts for sequencing depth.

        Normalizes by dividing sample counts using the given (sample) size
        factors. If no size factors are given, they are calculated using the
        median-of-ratios approach used by DESeq2.

        Parameters
        ----------
        size_factors : np.array
            Array of size factors, length should be equal to the number
            of samples.
        log2 : bool
            Whether to also log2-transform the normalized counts.

        Returns
        -------
        ExpressionMatrix
            ExpressionMatrix containing normalized counts.

        """

        with np.errstate(divide="ignore"):
            if size_factors is None:
                size_factors = self._estimate_size_factors(self._values)
            normalized = self._values.divide(size_factors, axis=1)

        if log2:
            normalized = np.log2(normalized + 1)

        return self._constructor(normalized)

    @staticmethod
    def _estimate_size_factors(counts):
        """Calculate size factors for DESeq's median-of-ratios normalization."""

        def _estimate_size_factors_col(counts, log_geo_means):
            log_counts = np.log(counts)
            mask = np.isfinite(log_geo_means) & (counts > 0)
            return np.exp(np.median((log_counts - log_geo_means)[mask]))

        with np.errstate(divide="ignore"):
            log_geo_means = np.mean(np.log(counts), axis=1)

            size_factors = np.apply_along_axis(
                _estimate_size_factors_col,
                axis=0,
                arr=counts,
                log_geo_means=log_geo_means)

        return size_factors



"""Tests for tree-related classes/functions."""

import pytest

from genopandas.core.tree import Interval, GenomicIntervalTree

# pylint: disable=redefined-outer-name,no-self-use


@pytest.fixture()
def simple_tuples():
    """Tuples for simple tree."""
    return [('2', 10, 20), ('2', 30, 40), ('3', 15, 25), ('3', 20, 25)]


@pytest.fixture()
def simple_tree(simple_tuples):
    """Simple tree containing a few intervals."""
    return GenomicIntervalTree.from_tuples(simple_tuples)


class TestGenomicIntervalTree(object):
    """Unit tests for the GenomicIntervalTree class."""

    def test_from_tuples(self, simple_tuples):
        """Build example tree from a set of tuples."""

        tree = GenomicIntervalTree.from_tuples(simple_tuples)

        assert len(tree) == 2
        assert Interval(10, 20) in tree['2']
        assert Interval(15, 25) in tree['3']

    def test_search(self, simple_tree):
        """Test search function."""

        # Check overlapping region.
        assert len(simple_tree.search('2', 5, 15)) == 1

        # Check region just outside interval (end is not inclusive).
        assert len(simple_tree.search('2', 5, 10)) == 0

        # Check regions on other chromosome.
        assert len(simple_tree.search('3', 15, 20)) == 1
        assert len(simple_tree.search('3', 20, 30)) == 2

        # Check region on non-existant chromosome.
        with pytest.raises(KeyError):
            simple_tree.search('4', 30, 40)

    def test_search_via_index(self, simple_tree):
        """Test search via index."""

        # Check overlapping region.
        assert len(simple_tree['2'].search(5, 15)) == 1

        # Check region just outside interval (end is not inclusive).
        assert len(simple_tree['2'].search(5, 10)) == 0

        # Check regions on other chromosome.
        assert len(simple_tree['3'].search(15, 20)) == 1
        assert len(simple_tree['3'].search(20, 30)) == 2

        # Check region on non-existant chromosome.
        with pytest.raises(KeyError):
            simple_tree['4'].search(30, 40)

    def test_intersection(self, simple_tree):
        """Test intersection."""

        tree_b = GenomicIntervalTree.from_tuples([('2', 10, 20)])
        intersection = simple_tree.intersection(tree_b)

        # Should drop unshared chromosome 3.
        assert len(intersection) == 1
        assert len(intersection['2']) == 1

        # Test operator.
        assert intersection == (tree_b & simple_tree)

    def test_union(self, simple_tree):
        """Test union."""

        tree_b = GenomicIntervalTree.from_tuples(
            [('3', 30, 40), ('4', 20, 30)])  # yapf: disable

        merged = simple_tree.union(tree_b)
        assert len(merged) == 3

        # CHeck presence of intervals.
        assert Interval(30, 40) in merged['3']
        assert Interval(20, 30) in merged['4']

        # Check original trees are unchanged.
        assert len(simple_tree) == 2
        assert len(tree_b) == 2

        # Test operator.
        assert simple_tree.union(tree_b) == (simple_tree | tree_b)

    def test_difference(self, simple_tree):
        """Test difference."""

        tree_b = GenomicIntervalTree.from_tuples(
            [('2', 10, 20), ('2', 30, 40), ('3', 5, 10)])  # yapf: disable
        diff = simple_tree.difference(tree_b)

        # Should not have dropped chromosome 2, but it should be empty..
        assert len(diff) == 2
        assert diff['2'].is_empty()

        # Chromosome 3 should be unchanged.
        assert diff['3'] == simple_tree['3']

        # Test operator.
        assert simple_tree.difference(tree_b) == (simple_tree - tree_b)



import functools
import itertools
import operator
import re

import numpy as np
import pandas as pd
from pandas.api.types import is_numeric_dtype
import toolz

from genopandas import plotting as gplot
from genopandas.util.pandas_ import DfWrapper

from .frame import GenomicDataFrame, GenomicSlice

RANGED_REGEX = r'(?P<chromosome>\w+):(?P<start>\d+)-(?P<end>\d+)'
POSITIONED_REGEX = r'(?P<chromosome>\w+):(?P<position>\d+)'


class AnnotatedMatrix(DfWrapper):
    """AnnotatedMatrix class.

    Annotated matrix classes respresent 2D numeric feature-by-sample matrices
    (with 'features' along the rows and samples along the columns), which can
    be annotated with optional sample_data and feature_data frames that
    describe the samples. The type of feature varies between different
    sub-classes, examples being genes (for gene expression matrices) and
    region-based bins (for copy-number data).

    This (base) class mainly contains a variety of methods for querying,
    subsetting and combining different annotation matrices. General plotting
    methods are also provided (``plot_heatmap``).

    Note that the class follows the feature-by-sample convention that is
    typically followed in biological packages, rather than the sample-by-feature
    orientation. This has the additional advantage of allowing more complex
    indices (such as a region-based MultiIndex) for the features, which are
    more difficult to use for DataFrame columns than for rows.

    Attributes
    ----------
    values : pd.DataFrame or AnnotatedMatrix
        Matrix values.
    sample_data : pd.DataFrame
        DataFrame containing sample annotations, whose index corresponds
        with the columns of the matrix.
    feature_data : pd.DataFrame
        DataFrame containing feature annotations, whose index corresponds
        with the rows of the matrix.

    """

    def __init__(self, values, sample_data=None, feature_data=None):
        if isinstance(values, AnnotatedMatrix):
            # Copy values from existing matrix (only copies sample/feature
            # data if these are not given explictly).
            sample_data = sample_data or values.sample_data
            feature_data = feature_data or values.feature_data
            values = values.values.copy()
        else:
            # Create empty annotations if none given.
            if sample_data is None:
                sample_data = pd.DataFrame({}, index=values.columns)

            if feature_data is None:
                feature_data = pd.DataFrame({}, index=values.index)

            # Check {sample,feature}_data.
            # assert (values.shape[1] == sample_data.shape[0]
            #         and all(values.columns == sample_data.index))

            # assert (values.shape[0] == feature_data.shape[0]
            #         and all(values.index == feature_data.index))

            # Check if all matrix columns are numeric.
            for col_name, col_values in values.items():
                if not is_numeric_dtype(col_values):
                    raise ValueError(
                        'Column {} is not numeric'.format(col_name))

        super().__init__(values)

        self._sample_data = sample_data.reindex(index=values.columns)
        self._feature_data = feature_data.reindex(index=values.index)

    def _constructor(self, values):
        """Constructor that attempts to build new instance
           from given values."""

        if isinstance(values, pd.DataFrame):
            return self.__class__(
                values.copy(),
                sample_data=self._sample_data,
                feature_data=self._feature_data)

        return values

    @property
    def feature_data(self):
        return self._feature_data

    @feature_data.setter
    def feature_data(self, value):
        value = value.reindex(index=self._values.index)
        self._feature_data = value

    @property
    def sample_data(self):
        return self._sample_data

    @sample_data.setter
    def sample_data(self, value):
        value = value.reindex(index=self._values.columns)
        self._sample_data = value

    @classmethod
    def from_csv(cls,
                 file_path,
                 sample_data=None,
                 feature_data=None,
                 sample_mapping=None,
                 feature_mapping=None,
                 drop_cols=None,
                 read_data_kws=None,
                 **kwargs):

        default_kwargs = {'index_col': 0}
        kwargs = toolz.merge(default_kwargs, kwargs)

        values = pd.read_csv(str(file_path), **kwargs)

        # If sample/feature_data are not dataframes, assume they are
        # file paths or objects and try to read from them.
        read_data_kws_default = {
            'sep': kwargs.pop('sep', None),
            'index_col': 0
        }
        read_data_kws = toolz.merge(read_data_kws_default, read_data_kws or {})

        if not (sample_data is None or isinstance(sample_data, pd.DataFrame)):
            sample_data = pd.read_csv(sample_data, **read_data_kws)

        if not (feature_data is None
                or isinstance(feature_data, pd.DataFrame)):
            feature_data = pd.read_csv(feature_data, **read_data_kws)

        values = cls._preprocess_values(
            values,
            sample_data=sample_data,
            feature_data=feature_data,
            sample_mapping=sample_mapping,
            feature_mapping=feature_mapping,
            drop_cols=drop_cols)

        return cls(values, sample_data=sample_data, feature_data=feature_data)

    @classmethod
    def _preprocess_values(cls,
                           values,
                           sample_data=None,
                           feature_data=None,
                           sample_mapping=None,
                           feature_mapping=None,
                           drop_cols=None):
        """Preprocesses matrix to match given sample/feature data."""

        # Drop extra columns (if needed).
        if drop_cols is not None:
            values = values.drop(drop_cols, axis=1)

        # Rename samples/features using mappings (if given).
        if sample_mapping is not None or feature_mapping is not None:
            values = values.rename(
                columns=sample_mapping, index=feature_mapping)

        # Reorder values to match annotations.
        sample_order = None if sample_data is None else sample_data.index
        feat_order = None if feature_data is None else feature_data.index

        values = values.reindex(
            columns=sample_order, index=feat_order, copy=False)

        return values

    def to_csv(self,
               file_path,
               sample_data_path=None,
               feature_data_path=None,
               **kwargs):
        """Writes matrix values to a csv file, using pandas' to_csv method."""

        # Write matrix values.
        self._values.to_csv(file_path, **kwargs)

        # Write sample/feature data if paths given.
        if sample_data_path is not None:
            self._sample_data.to_csv(
                sample_data_path, sep=kwargs.pop('sep', None), index=True)

        if feature_data_path is not None:
            self._feature_data.to_csv(
                feature_data_path, sep=kwargs.pop('sep', None), index=True)

    def rename(self, index=None, columns=None):
        """Rename samples/features in the matrix."""

        renamed = self._values.rename(index=index, columns=columns)

        if index is not None:
            feature_data = self._feature_data.rename(index=index)
        else:
            feature_data = self._feature_data

        if columns is not None:
            sample_data = self._sample_data.rename(index=columns)
        else:
            sample_data = self._sample_data

        return self.__class__(
            renamed, feature_data=feature_data, sample_data=sample_data)

    def melt(self,
             with_sample_data=False,
             with_feature_data=False,
             value_name='value'):
        """Melts values into 'tidy' format, optionally including annotation."""

        feat_col = self._feature_data.index.name or 'feature'
        sample_col = self._sample_data.index.name or 'sample'

        values_long = pd.melt(
            self._values.rename_axis(feat_col).reset_index(),
            id_vars=feat_col,
            var_name=sample_col,
            value_name=value_name)

        if with_sample_data and self._sample_data.shape[1] > 0:
            sample_data = (self._sample_data.rename_axis(sample_col)
                           .reset_index())

            values_long = pd.merge(
                values_long, sample_data, how='left', on=sample_col)

        if with_feature_data and self._feature_data.shape[1] > 0:
            feature_data = (self._feature_data.rename_axis(feat_col)
                            .reset_index())

            # Merge with annotation.
            values_long = pd.merge(
                values_long, feature_data, how='left', on=feat_col)

        return values_long

    def query_samples(self, expr):
        """Subsets samples in matrix by querying sample_data with expression.

        Similar to the pandas ``query`` method, this method queries the sample
        data of the matrix with the given boolean expression. Any samples for
        which the expression evaluates to True are returned in the resulting
        AnnotatedMatrix.

        Parameters
        ----------
        expr : str
            The query string to evaluate. You can refer to variables in the
            environment by prefixing them with an ‘@’ character like @a + b.

        Returns
        -------
        AnnotatedMatrix
            Subsetted matrix, containing only the samples for which ``expr``
            evaluates to True.

        """

        sample_data = self._sample_data.query(expr)
        values = self._values.reindex(columns=sample_data.index)

        return self.__class__(
            values, sample_data=sample_data, feature_data=self._feature_data)

    def dropna_samples(self, subset=None, how='any', thresh=None):
        """Drops samples with NAs in sample_data."""

        sample_data = self._sample_data.dropna(
            subset=subset, how=how, thresh=thresh)
        values = self._values.reindex(columns=sample_data.index)

        return self.__class__(
            values, sample_data=sample_data, feature_data=self._feature_data)

    def __eq__(self, other):
        if not isinstance(other, AnnotatedMatrix):
            return False
        return all(self.values == other.values) and \
            all(self.sample_data == other.sample_data) and \
            all(self.feature_data == other.feature_data)

    def plot_heatmap(
            self,
            cmap='RdBu_r',
            sample_cols=None,
            sample_colors=None,
            feature_cols=None,
            feature_colors=None,
            metric='euclidean',
            method='complete',
            transpose=False,
            # legend_kws=None,
            **kwargs):
        """Plots clustered heatmap of matrix values."""

        import matplotlib.pyplot as plt
        import seaborn as sns

        if sample_cols is not None:
            sample_annot, _ = gplot.color_annotation(
                self._sample_data[sample_cols], colors=sample_colors)
        else:
            sample_annot, _ = None, None

        if feature_cols is not None:
            feature_annot, _ = gplot.color_annotation(
                self._feature_data[feature_cols], colors=feature_colors)
        else:
            feature_annot, _ = None, None

        clustermap_kws = dict(kwargs)

        if transpose:
            values = self._values.T
            clustermap_kws['row_colors'] = sample_annot
            clustermap_kws['col_colors'] = feature_annot
            xlabel, ylabel = 'Features', 'Samples'
        else:
            values = self._values
            clustermap_kws['col_colors'] = sample_annot
            clustermap_kws['row_colors'] = feature_annot
            xlabel, ylabel = 'Samples', 'Features'

        cm = sns.clustermap(
            values, cmap=cmap, metric=metric, method=method, **clustermap_kws)

        plt.setp(cm.ax_heatmap.get_yticklabels(), rotation=0)

        cm.ax_heatmap.set_xlabel(xlabel)
        cm.ax_heatmap.set_ylabel(ylabel)

        # TODO: handle legend drawing.
        #if annot_cmap is not None:
        #    draw_legends(cm, annot_cmap, **(legend_kws or {}))

        return cm

    def pca(self,
            n_components=None,
            axis='columns',
            transform=False,
            with_annotation=False):
        """Performs PCA on matrix."""

        try:
            from sklearn.decomposition import PCA
        except ImportError:
            raise ImportError('Scikit-learn must be installed to '
                              'perform PCA analyses')

        # Fit PCA and transform expression.
        pca = PCA(n_components=n_components)

        if axis in {1, 'columns', 'samples'}:
            values = self._values.T
            annotation = self._sample_data
        elif axis in {0, 'index', 'features'}:
            values = self._values
            annotation = self._feature_data
        else:
            raise ValueError('Unknown value for axis')

        pca.fit(values.values)

        if transform:
            transformed = pca.transform(values.values)

            n_components = transformed.shape[1]

            transformed = pd.DataFrame(
                transformed,
                columns=['pca_{}'.format(i + 1) for i in range(n_components)],
                index=values.index)

            if with_annotation:
                transformed = pd.concat([transformed, annotation], axis=1)

            return pca, transformed
        else:
            return pca

    def plot_pca(self, components=(1, 2), axis='columns', ax=None, **kwargs):
        """Plots PCA of samples."""

        pca, transformed = self.pca(
            n_components=max(components),
            axis=axis,
            transform=True,
            with_annotation=True)

        # Draw using lmplot.
        pca_x, pca_y = ['pca_{}'.format(c) for c in components]
        ax = gplot.scatter_plot(
            data=transformed, x=pca_x, y=pca_y, ax=ax, **kwargs)

        var = pca.explained_variance_ratio_[components[0] - 1] * 100
        ax.set_xlabel('Component {} ({:3.1f}%)'.format(components[0], var))

        var = pca.explained_variance_ratio_[components[1] - 1] * 100
        ax.set_ylabel('Component {} ({:3.1f}%)'.format(components[1], var))

        return ax

    def plot_pca_variance(self, n_components=None, axis='columns', ax=None):
        """Plots variance explained by PCA components."""

        import matplotlib.pyplot as plt
        import seaborn as sns

        pca = self.pca(n_components=n_components, axis=axis, transform=False)

        if ax is None:
            _, ax = plt.subplots()

        x = np.arange(pca.n_components_) + 1
        y = pca.explained_variance_ratio_
        ax.plot(x[:len(y)], y)

        ax.set_xlabel('Component')
        ax.set_ylabel('Explained variance')
        sns.despine(ax=ax)

        return ax

    def plot_feature(self, feature, group=None, kind='box', ax=None, **kwargs):
        """Plots distribution of expression for given feature."""

        import seaborn as sns

        if group is not None and self._sample_data.shape[1] == 0:
            raise ValueError('Grouping not possible without sample data')

        # Determine plot type.
        plot_funcs = {
            'box': sns.boxplot,
            'swarm': sns.swarmplot,
            'violin': sns.violinplot
        }

        try:
            plot_func = plot_funcs[kind]
        except KeyError:
            raise ValueError('Unknown plot type {!r}'.format(kind))

        # Assemble plot data (sample_data + expression values).
        values = self._values.loc[feature].to_frame(name='value')
        plot_data = pd.concat([values, self._sample_data], axis=1)

        # Plot expression.
        ax = plot_func(data=plot_data, x=group, y='value', ax=ax, **kwargs)

        ax.set_title(feature)
        ax.set_ylabel('Value')

        return ax

    @classmethod
    def concat(cls, matrices, axis):
        """Concatenates matrices along given axis."""

        # Collect value/sample/feature data.
        tuples = ((mat.values, mat.sample_data, mat.feature_data)
                  for mat in matrices)
        value_list, sample_list, feat_list = zip(*tuples)

        # Merge values.
        values = pd.concat(value_list, axis=axis)

        # Merge sample/feature data.
        if axis == 'index' or axis == 0:
            sample_data = pd.concat(sample_list, axis='columns')
            feature_data = pd.concat(feat_list, axis='index')
        elif axis == 'columns' or axis == 1:
            sample_data = pd.concat(sample_list, axis='index')
            feature_data = pd.concat(feat_list, axis='columns')
        else:
            raise ValueError('Unknown value for axis')

        return cls(values, sample_data=sample_data, feature_data=feature_data)

    def drop_duplicate_indices(self, axis='index', keep='first'):
        """Drops duplicate indices along given axis."""

        if axis == 'index':
            mask = ~self._values.index.duplicated(keep=keep)
            values = self._values.loc[mask]

            sample_data = self._sample_data
            feature_data = self._feature_data.loc[mask]
        elif axis == 'columns':
            mask = ~self._values.columns.duplicated(keep=keep)
            values = self._values.loc[:, mask]

            sample_data = self._sample_data.loc[mask]
            feature_data = self._feature_data
        else:
            raise ValueError('Unknown value for axis')

        return self.__class__(
            values.copy(), sample_data=sample_data, feature_data=feature_data)


class GenomicMatrix(AnnotatedMatrix):
    """Class respresenting matrices indexed by genomic positions."""

    def __init__(self, values, sample_data=None, feature_data=None):
        if not isinstance(values, GenomicDataFrame):
            values = GenomicDataFrame(values)

        super().__init__(
            values, sample_data=sample_data, feature_data=feature_data)

    @classmethod
    def from_df(cls, values, chrom_lengths=None, **kwargs):
        """Constructs a genomic matrix from the given DataFrame."""

        if not isinstance(values, GenomicDataFrame):
            values = GenomicDataFrame.from_df(
                values, chrom_lengths=chrom_lengths)

        return cls(values, **kwargs)

    @classmethod
    def from_csv(cls,
                 file_path,
                 index_col,
                 sample_data=None,
                 feature_data=None,
                 sample_mapping=None,
                 feature_mapping=None,
                 drop_cols=None,
                 chrom_lengths=None,
                 read_data_kws=None,
                 **kwargs):
        """Reads values from a csv file."""

        if not 2 <= len(index_col) <= 3:
            raise ValueError('index_col should contain 2 entries'
                             ' (for positioned data or 3 entries'
                             ' (for ranged data)')

        default_dtype = {index_col[0]: str}
        dtype = toolz.merge(default_dtype, kwargs.pop('dtype', {}))

        values = pd.read_csv(file_path, dtype=dtype, **kwargs)
        values = values.set_index(index_col)

        # If sample/feature_data are not dataframes, assume they are
        # file paths or objects and try to read from them.
        read_data_kws_default = {
            'sep': kwargs.pop('sep', None),
            'index_col': 0
        }
        read_data_kws = toolz.merge(read_data_kws_default, read_data_kws or {})

        if not (sample_data is None or isinstance(sample_data, pd.DataFrame)):
            sample_data = pd.read_csv(sample_data, **read_data_kws)

        if not (feature_data is None
                or isinstance(feature_data, pd.DataFrame)):
            feature_data = pd.read_csv(feature_data, **read_data_kws)

        values = cls._preprocess_values(
            values,
            sample_data=sample_data,
            feature_data=feature_data,
            sample_mapping=sample_mapping,
            feature_mapping=feature_mapping,
            drop_cols=drop_cols)

        return cls.from_df(
            values,
            sample_data=sample_data,
            feature_data=feature_data,
            chrom_lengths=chrom_lengths)

    @classmethod
    def from_csv_condensed(cls,
                           file_path,
                           index_col=0,
                           sample_data=None,
                           feature_data=None,
                           sample_mapping=None,
                           feature_mapping=None,
                           drop_cols=None,
                           chrom_lengths=None,
                           index_regex=RANGED_REGEX,
                           is_one_based=False,
                           is_inclusive=False,
                           read_data_kws=None,
                           **kwargs):
        """Reads values from a csv file with a condensed index."""

        values = pd.read_csv(file_path, index_col=index_col, **kwargs)

        values.index = cls._expand_condensed_index(
            values.index,
            index_regex,
            is_one_based=is_one_based,
            is_inclusive=is_inclusive)

        # If sample/feature_data are not dataframes, assume they are
        # file paths or objects and try to read from them.
        read_data_kws_default = {
            'sep': kwargs.pop('sep', None),
            'index_col': 0
        }
        read_data_kws = toolz.merge(read_data_kws_default, read_data_kws or {})

        if not (sample_data is None or isinstance(sample_data, pd.DataFrame)):
            sample_data = pd.read_csv(sample_data, **read_data_kws)

        if not (feature_data is None
                or isinstance(feature_data, pd.DataFrame)):
            feature_data = pd.read_csv(feature_data, **read_data_kws)

        values = cls._preprocess_values(
            values,
            sample_data=sample_data,
            feature_data=feature_data,
            sample_mapping=sample_mapping,
            feature_mapping=feature_mapping,
            drop_cols=drop_cols)

        return cls.from_df(
            values,
            sample_data=sample_data,
            feature_data=feature_data,
            chrom_lengths=chrom_lengths)

    @classmethod
    def _expand_condensed_index(cls,
                                index,
                                regex_expr,
                                is_one_based=False,
                                is_inclusive=False):
        """Expands condensed index into a MultiIndex."""

        # Parse entries.
        regex = re.compile(regex_expr)
        group_dicts = (regex.match(el).groupdict() for el in index)

        # Extract chromosome, start, end positions.
        if regex.groups == 3:
            tups = ((grp['chromosome'], int(grp['start']), int(grp['end']))
                    for grp in group_dicts)
            chrom, starts, ends = zip(*tups)
        elif regex.groups == 2:
            tups = ((grp['chromosome'], int(grp['position']))
                    for grp in group_dicts)
            chrom, starts = zip(*tups)
            ends = None
        else:
            raise ValueError('Regex should have two or three groups '
                             '(for positioned/ranged data, respectively)')

        # Correct for one-base and inclusive-ness to match Python conventions.
        starts = np.array(starts)

        if is_one_based:
            starts -= 1

        if ends is not None and is_inclusive:
            ends = np.array(ends)
            ends += 1

        # Build index.
        if ends is None:
            index = pd.MultiIndex.from_arrays(
                [chrom, starts], names=['chromosome', 'position'])
        else:
            index = pd.MultiIndex.from_arrays(
                [chrom, starts, ends], names=['chromosome', 'start', 'end'])

        return index

    @property
    def gloc(self):
        """Genomic-position indexer.

        Used to select rows from the matrix by their genomic position.
        Interface is the same as for the GenomicDataFrame gloc property
        (which this method delegates to).
        """

        return GLocWrapper(self._values.gloc, self._gloc_constructor)

    def _gloc_constructor(self, values):
        """Constructor that attempts to build new instance
           from given values."""

        if isinstance(values, GenomicDataFrame):
            sample_data = self._sample_data.reindex(index=values.columns)
            feature_data = self._feature_data.reindex(index=values.index)

            return self.__class__(
                values.copy(),
                sample_data=sample_data,
                feature_data=feature_data)

        return values

    def expand(self):
        """Expands matrix to include values from missing bins.

        Assumes rows are regularly spaced with a fixed bin size.
        """

        expanded = self._expand(self._values)
        feature_data = self._feature_data.reindex(index=expanded.index)

        return self.__class__(
            expanded, sample_data=self._sample_data, feature_data=feature_data)

    @staticmethod
    def _expand(values):
        def _bin_indices(grp, bin_size):
            chrom = grp.index[0][0]

            start = grp.index.get_level_values(1).min()
            end = grp.index.get_level_values(2).max()

            bins = np.arange(start, end + 1, step=bin_size)

            return zip(itertools.cycle([chrom]), bins[:-1], bins[1:])

        bin_size = values.index[0][2] - values.index[0][1]

        # TODO: Warn if bin_size is 1? (Probably positioned data).

        # Check inferred bin size.
        starts = values.index.get_level_values(1)
        ends = values.index.get_level_values(2)
        diffs = ends - starts

        if not all(diffs == bin_size):
            raise ValueError('Bins do not match inferred bin size')

        # Check if following bins match inferred bin size.
        if not all(np.mod(np.diff(starts), bin_size) == 0):
            raise ValueError('Following bins do not match inferred bin size')

        indices = list(
            itertools.chain.from_iterable(
                _bin_indices(grp, bin_size=bin_size)
                for _, grp in values.groupby(level=0)))

        return values.reindex(index=indices)

    def impute(self, window=11, min_probes=5, expand=True):
        """Imputes nan values from neighboring bins."""

        if expand:
            values = self._expand(self._values)
        else:
            values = self._values

        # Calculate median value within window (allowing for
        # window - min_probes number of NAs within the window).
        rolling = values.rolling(
            window=window, min_periods=min_probes, center=True)
        avg_values = rolling.median()

        # Copy over values for null rows for the imputation.
        imputed = values.copy()

        mask = imputed.isnull().all(axis=1)
        imputed.loc[mask] = avg_values.loc[mask]

        # Match feature data to new values.
        feature_data = self._feature_data.reindex(index=imputed.index)

        return self.__class__(
            imputed, sample_data=self._sample_data, feature_data=feature_data)

    def resample(self, bin_size, start=None, agg='mean'):
        """Resamples values at given interval by binning."""

        # Perform resampling per chromosome.
        resampled = pd.concat(
            (self._resample_chromosome(
                grp, bin_size=bin_size, agg=agg, start=start)
             for _, grp in self._values.groupby(level=0)),
            axis=0)  # yapf: disable

        # Restore original index order.
        resampled = resampled.reindex(self._values.gloc.chromosomes, level=0)

        return self.__class__(
            GenomicDataFrame(
                resampled, chrom_lengths=self._values.chromosome_lengths),
            sample_data=self._sample_data)

    @staticmethod
    def _resample_chromosome(values, bin_size, start=None, agg='mean'):
        # Bin rows by their centre positions.
        starts = values.index.get_level_values(1)
        ends = values.index.get_level_values(2)

        positions = (starts + ends) // 2

        range_start = starts.min() if start is None else start
        range_end = ends.max() + bin_size

        bins = np.arange(range_start, range_end, bin_size)

        if len(bins) < 2:
            raise ValueError('No bins in range ({}, {}) with bin_size {}'.
                             format(range_start, ends.max(), bin_size))

        binned = pd.cut(positions, bins=bins)

        # Resample.
        resampled = values.groupby(binned).agg(agg)
        resampled.index = pd.MultiIndex.from_arrays(
            [[values.index[0][0]] * (len(bins) - 1), bins[:-1], bins[1:]],
            names=values.index.names)

        return resampled

    def rename_chromosomes(self, mapping):
        """Returns copy of matrix with renamed chromosomes."""

        return self.__class__(
            values=self._values.rename_chromosomes(mapping),
            sample_data=self.sample_data,
            feature_data=self.feature_data)

    def annotate(self, features, feature_id='gene_id'):
        """Annotates values for given features."""

        # Calculate calls.
        get_id = operator.attrgetter(feature_id)

        annotated_calls = {}
        for feature in features.itertuples():
            try:
                chrom, start, end = feature.Index
                overlap = self._values.gloc.search(chrom, start, end)

                annotated_calls[get_id(feature)] = overlap.median()
            except KeyError:
                pass

        # Assemble into dataframe.
        annotated = pd.DataFrame.from_records(annotated_calls).T
        annotated.index.name = feature_id

        return AnnotatedMatrix(annotated, sample_data=self._sample_data)

    def plot_sample(self, sample, ax=None, **kwargs):
        """Plots values for given sample along genomic axis."""
        ax = gplot.genomic_scatter_plot(
            self._values, y=sample, ax=ax, **kwargs)
        return ax

    def plot_heatmap(self,
                     cmap='RdBu_r',
                     sample_cols=None,
                     sample_colors=None,
                     metric='euclidean',
                     method='complete',
                     transpose=True,
                     cluster=True,
                     **kwargs):
        """Plots heatmap of gene expression over samples."""

        if 'row_cluster' in kwargs or 'col_cluster' in kwargs:
            raise ValueError(
                'GenomicMatrices only supports clustering by samples. '
                'Use the \'cluster\' argument to specify whether '
                'clustering should be performed.')

        if cluster:
            from scipy.spatial.distance import pdist
            from scipy.cluster.hierarchy import linkage

            # Do clustering on matrix with only finite values.
            values_clust = self._values.replace([np.inf, -np.inf], np.nan)
            values_clust = values_clust.dropna()

            dist = pdist(values_clust.T, metric=metric)
            sample_linkage = linkage(dist, method=method)
        else:
            sample_linkage = None

        # Draw heatmap.
        heatmap_kws = dict(kwargs)

        if transpose:
            heatmap_kws.update({
                'row_cluster': sample_linkage is not None,
                'row_linkage': sample_linkage,
                'col_cluster': False
            })
        else:
            heatmap_kws.update({
                'col_cluster': sample_linkage is not None,
                'col_linkage': sample_linkage,
                'row_cluster': False
            })

        cm = super().plot_heatmap(
            sample_cols=sample_cols,
            sample_colors=sample_colors,
            cmap=cmap,
            metric=metric,
            method=method,
            transpose=transpose,
            **heatmap_kws)

        self._style_heatmap(cm, transpose=transpose)

        return cm

    def _style_heatmap(self, cm, transpose):
        chrom_breaks = self._values.groupby(level=0).size().cumsum()

        chrom_labels = self._values.gloc.chromosomes
        chrom_label_pos = np.concatenate([[0], chrom_breaks])
        chrom_label_pos = (chrom_label_pos[:-1] + chrom_label_pos[1:]) / 2

        if transpose:
            cm.ax_heatmap.set_xticks([])

            for loc in chrom_breaks[:-1]:
                cm.ax_heatmap.axvline(loc, color='grey', lw=1)

            cm.ax_heatmap.set_xticks(chrom_label_pos)
            cm.ax_heatmap.set_xticklabels(chrom_labels, rotation=0)

            cm.ax_heatmap.set_xlabel('Genomic position')
            cm.ax_heatmap.set_ylabel('Samples')
        else:
            cm.ax_heatmap.set_yticks([])

            for loc in chrom_breaks[:-1]:
                cm.ax_heatmap.axhline(loc, color='grey', lw=1)

            cm.ax_heatmap.set_yticks(chrom_label_pos)
            cm.ax_heatmap.set_yticklabels(chrom_labels, rotation=0)

            cm.ax_heatmap.set_xlabel('Samples')
            cm.ax_heatmap.set_ylabel('Genomic position')

        return cm


class GLocWrapper(object):
    """Wrapper class that wraps gloc indexer from given object."""

    def __init__(self, gloc, constructor):
        self._gloc = gloc
        self._constructor = constructor

    def __getattr__(self, name):
        attr = getattr(self._gloc, name)

        if callable(attr):
            return self._wrap_function(attr)

        return attr

    def __getitem__(self, item):
        result = self._gloc[item]

        if isinstance(result, GenomicSlice):
            result = GLocSliceWrapper(
                self._gloc, chromosome=item, constructor=self._constructor)
        else:
            result = self._constructor(result)

        return result

    def _wrap_function(self, func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            """Wrapper that calls _constructor on returned result."""
            result = func(*args, **kwargs)
            return self._constructor(result)

        return wrapper


class GLocSliceWrapper(object):
    """Wrapper class that wraps slice from gloc indexer on given object."""

    def __init__(self, gloc, chromosome, constructor):
        self._gloc = gloc
        self._chromosome = chromosome
        self._constructor = constructor

    def __getitem__(self, item):
        result = self._gloc[self._chromosome][item]
        return self._constructor(result)



import os
from pathlib import Path

import numpy as np
import random as rand

pytest_plugins = ['helpers_namespace']

import pytest

BASE_DIR = Path(__file__).parent


@pytest.helpers.register
def data_path(relative_path, relative_to=None):
    """Returns data path to test file."""

    if relative_to is None:
        # Use BASE_DIR as default.
        relative_to = BASE_DIR
    elif not isinstance(relative_to, Path):
        # Ensure relative_to is a Path.
        relative_to = Path(relative_to)

    # If relative_to is not a path, move up one level.
    if not relative_to.is_dir():
        relative_to = relative_to.parent

    return relative_to / 'data' / relative_path


@pytest.fixture
def random():
    rand.seed(0)
    np.random.seed(0)



"""Tree-related functions/classes."""

import collections
import itertools
import operator

# pylint: disable=unused-import
from intervaltree import IntervalTree, Interval

# pylint: enable=unused-import


class GenomicIntervalTree(collections.MutableMapping):
    """Datastructure for efficiently accessing genomic objects by position."""

    def __init__(self, *args, **kwargs):
        self._trees = dict()
        self.update(dict(*args, **kwargs))

    def __getitem__(self, key):
        return self._trees[key]

    def __setitem__(self, key, value):
        self._trees[key] = value

    def __delitem__(self, key):
        del self._trees[key]

    def __iter__(self):
        return iter(self._trees)

    def __len__(self):
        return len(self._trees)

    def __repr__(self):
        args = ', '.join('{!r}: {}'.format(k, v) for k, v in self.items())
        return 'GenomicIntervalTree(**{' + args + '})'

    # pylint: disable=too-many-arguments
    def search(self,
               chromosome,
               start,
               end=None,
               strict_left=False,
               strict_right=False):
        """Searches the tree for objects within given range."""

        overlap = self[chromosome].search(start, end)

        if strict_left:
            overlap = (int_ for int_ in overlap if int_.begin >= start)

        if strict_right:
            overlap = (int_ for int_ in overlap if int_.end <= end)

        return set(overlap)

    # pylint: enable=too-many-arguments

    def is_empty(self):
        """Returns True if tree is empty."""

        if len(self) == 0:
            return True

        return all(tree.isempty() for tree in self._trees.values())

    @classmethod
    def from_tuples(cls, tuples):
        """Builds an instance from tuples.

        Assumes tuples are sorted by chromosome.
        """

        # Group by chromosome.
        tuples = sorted(tuples, key=operator.itemgetter(0))
        groups = itertools.groupby(tuples, key=operator.itemgetter(0))

        # Build trees.
        trees = {}
        for chrom, group in groups:
            trees[chrom] = IntervalTree.from_tuples(tup[1:] for tup in group)

        return cls(trees)

    def intersection(self, other):
        """Returns a new tree of all intervals common to both self and other."""

        shared = set(self.keys()) & set(other.keys())
        return self.__class__(
            {k: self[k].intersection(other[k])
             for k in shared})

    def union(self, other):
        """Returns a new tree, comprising all intervals from self and other."""

        merged = {chrom: tree.copy() for chrom, tree in self._trees.items()}

        for chrom, tree in other.items():
            if chrom in merged:
                merged[chrom] = merged[chrom].union(tree)
            else:
                merged[chrom] = tree.copy()

        return self.__class__(merged)

    def difference(self, other):
        """Returns a new tree, comprising all intervals in self but not
           in other."""

        diff = {}

        for chrom, tree in self.items():
            if chrom in other:
                diff[chrom] = tree.difference(other[chrom])
            else:
                diff[chrom] = tree.copy()

        return self.__class__(diff)

    def __or__(self, other):
        return self.union(other)

    def __and__(self, other):
        return self.intersection(other)

    def __sub__(self, other):
        return self.difference(other)



from .frame import GenomicDataFrame, GenomicIndexer
from .matrix import AnnotatedMatrix, GenomicMatrix
from .tree import Interval, IntervalTree, GenomicIntervalTree

__all__ = [
    'GenomicDataFrame', 'AnnotatedMatrix', 'GenomicMatrix', 'Interval',
    'IntervalTree', 'GenomicIntervalTree'
]



"""Module containing functions for plotting data along a genomic axis."""

from itertools import chain

import numpy as np
import pandas as pd
import toolz

from .base import scatter_plot, step_plot, apply_palette


def genomic_scatter_plot(data,
                         y,
                         hue=None,
                         hue_order=None,
                         palette=None,
                         color=None,
                         chromosomes=None,
                         legend=True,
                         legend_kws=None,
                         ax=None,
                         style_axis=True,
                         **kwargs):
    """Plots genomic data along a chromosomal axis.

    Parameters
    ----------
    data : GenomicDataFrame
        Genomic data to plot.
    y, hue : str
        Columns to use for plotting. ``y`` determines what is drawn on the
        y-axis. If given, ``hue`` points are colored according to the
        (categorical) values of the respective column. If hue == 'chromosome'
        points are colored per chromosome.
    hue_order : List[str]
        Order to plot the categorical hue levels in.
    palette : List[str] or Dict[Any, str]
        Colors to use for the different levels of the hue variable. Can either
        be a dictionary mapping values to specific colors, or a list of colors
        to use.
    color : matplotlib color
        Color to use for all elements. Overrides palette if given.
    chromosomes: List[str]
        List of chromosomes to plot. Can be used to select a subset of
        chromosomes or to specify a specific order.
    legend : bool
        Whether to draw a legend for the different hue levels.
        (Only used if hue is given.)
    legend_kws : Dict[str, Any]
        Dictionary of additional keyword arguments to pass to ax.legend
        when drawing the legend.
    ax : AxesSubplot
        Axis to use for drawing.
    kwargs : Dict[str, Any]
        Other keyword arguments are passed through to ``ax.plot`` at draw time.

    Returns
    -------
    AxesSubplot
        Axis on which the data was drawn.

    """

    if chromosomes is not None:
        data = data.gloc[chromosomes]

    # Assemble plot data.
    plot_data = pd.DataFrame({
        'chromosome': data.gloc.chromosome.values,
        'position': data.gloc.position_offset.values,
        'y': data[y].values
    })  # yapf: disable

    if hue is not None and hue not in plot_data:
        plot_data[hue] = data[hue]

    # Order hue by data chromosome order if hue == "chromosome"
    # and no specific order is given.
    if hue == 'chromosome' and hue_order is None:
        hue_order = data.gloc.chromosomes

    # Plot using scatter.
    default_plot_kws = {'markersize': 1}
    plot_kws = toolz.merge(default_plot_kws, kwargs)

    ax = scatter_plot(
        data=plot_data,
        x='position',
        y='y',
        hue=hue,
        hue_order=hue_order,
        palette=palette,
        color=color,
        legend=legend,
        legend_kws=legend_kws,
        ax=ax,
        **plot_kws)

    if style_axis:
        # Style axis.
        _draw_dividers(data.gloc.chromosome_offsets, ax=ax)

        ax.set_xlabel('Chromosome')
        ax.set_ylabel(y)

    return ax


def _draw_dividers(chrom_offsets, ax):
    """Draws chromosome dividers at offsets to indicate chromosomal boundaries.

    The chrom_offsets argument is expected to include _END_ marker (which is
    included by default in GenomicDataFrames).

    Parameters
    ----------
    chrom_offsets : Dict[str, int]
        Position offsets at which to draw boundaries for the
        respective chromosomes.
    ax : AxesSubplot
        Axis to use for drawing.
    """

    positions = np.array(list(chrom_offsets.values()))

    # Draw dividers.
    for loc in positions[1:-1]:
        ax.axvline(loc, color='grey', lw=0.5, zorder=5)

    # Draw xtick labels.
    ax.set_xticks((positions[:-1] + positions[1:]) / 2)
    ax.set_xticklabels(chrom_offsets.keys())

    # Set xlim to boundaries.
    ax.set_xlim(0, chrom_offsets['_END_'])


def genomic_step_plot(data,
                      y,
                      hue=None,
                      hue_order=None,
                      palette=None,
                      color=None,
                      chromosomes=None,
                      legend=True,
                      legend_kws=None,
                      ax=None,
                      style_axis=True,
                      **kwargs):

    if chromosomes is not None:
        data = data.gloc[chromosomes]

    # We need to include both start/end positions in the dataframe.
    # To do so, we basically create two copies of the original df
    # (one with start, one with end positions), concat these two frames
    # and then sort the concatenated frame by original index and position.

    # Create initial frame (with start positions).
    plot_data = pd.DataFrame({
        'chromosome': data.gloc.chromosome.values,
        'position': data.gloc.start_offset.values,
        'y': data[y].values,
        'index': np.arange(len(data[y]))
    })

    if hue is not None:
        plot_data[hue] = data[hue]

    # Merge with copy containing end positions.
    plot_data = pd.concat(
        [plot_data,
         plot_data.assign(position=data.gloc.end_offset.values)],
        axis=0)

    # Sort by original row order.
    plot_data = plot_data.sort_values(by=['index', 'position'])
    plot_data = plot_data.drop('index')

    # Order hue by data chromosome order if hue == "chromosome" and
    # no specific order is given.
    if hue == 'chromosome' and hue_order is None:
        hue_order = data.gloc.chromosomes

    # Plot using step.
    default_step_kws = {'where': 'post'}
    step_kws = toolz.merge(default_step_kws, kwargs)

    ax = step_plot(
        data=plot_data,
        x='position',
        y='y',
        hue=hue,
        hue_order=hue_order,
        palette=palette,
        color=color,
        legend=legend,
        legend_kws=legend_kws,
        ax=ax,
        **step_kws)

    if style_axis:
        # Style axis.
        _draw_dividers(data.gloc.chromosome_offsets, ax=ax)

        ax.set_xlabel('Chromosome')
        ax.set_ylabel(y)

    return ax


def genomic_region_plot(data,
                        y=None,
                        hue=None,
                        hue_order=None,
                        palette=None,
                        color=None,
                        chromosomes=None,
                        ax=None,
                        style_axis=True,
                        **kwargs):
    """Plots highlighted regions along a genomic axis.

    Parameters
    ----------
    data : pandas.DataFrame
        Tidy ('long-form'') dataframe where each column is a variable and
        each row is an observation. Should contain specified
        {chrom,start,end}_col columns (which default to 'chromosome', 'start'
        and 'end', respectively).
    hue : str
        Column to color the data points by.
    hue_order:
        Order to plot the categorical levels in, otherwise the levels
        are inferred from the data objects.
    palette:
        Colors to use for the different levels of the hue variable.
    color:
        Color for all of the elements. Only used when hue is not specified.
    chromosomes: List[str]
        List of chromosomes to plot. Can be used to select a subset of
        chromosomes or to specify a specific order.
    ax : matplotlib Axes, optional
        Axes object to draw the plot onto.
    style_axis : bool
        Whether to style axes with dividers, labels etc. If False, leaves the
        axis unchanged. Useful for combining with other functions
        (such as plot_genomic) to avoid double drawing of annotations.
    kwargs : Dict[str, Any]
        Other keyword arguments are passed through to ``PatchCollection``
        (when drawing with y values) or ``ax.axvspan`` (when drawing without
        y values) at draw time.

    Returns
    -------
    matplotlib.Axes
        The Axes object containing the plot.

    """

    from matplotlib import pyplot as plt

    if chromosomes is not None:
        data = data.gloc[chromosomes]

    # Default axes.
    if ax is None:
        _, ax = plt.subplots()

    # Assemble plot data.
    plot_data = pd.DataFrame({
        'chromosome': data.gloc.chromosome.values,
        'start': data.gloc.start_offset.values,
        'end': data.gloc.end_offset.values
    })

    if y is not None:
        plot_data['value'] = data[y].values
        draw_func = _draw_region_patches
    else:
        draw_func = _draw_region_spans

    if hue is None:
        draw_func(plot_data, ax=ax, color=color, **kwargs)
    else:
        plot_data['hue'] = data[hue]

        plot_data = plot_data.assign(_color=apply_palette(
            plot_data[hue], palette, order=hue_order))

        for (_, color), grp in plot_data.groupby(['hue', '_color']):
            draw_func(grp, ax=ax, color=color, **kwargs)

    if style_axis:
        # Style axis.

        _draw_dividers(data.gloc.chromosome_offsets, ax=ax)
        ax.set_xlabel('Chromosome')

        if y is not None:
            ax.set_ylabel(y)

    # TODO: Set ylim?

    return ax


def _draw_region_patches(grp, ax, color=None, **kwargs):
    from matplotlib import patches as mpl_patches
    from matplotlib import collections as mpl_collections

    grp = grp.assign(width=grp['end'] - grp['start'])

    patches = mpl_collections.PatchCollection(
        (mpl_patches.Rectangle(
            xy=(tup.start, 0), width=tup.width, height=tup.value)
         for tup in grp.itertuples()),
        facecolor=color,
        edgecolor=color,
        **kwargs)

    ax.add_collection(patches)


def _draw_region_spans(grp, ax, color=None, **kwargs):
    for tup in grp.itertuples():
        ax.axvspan(tup.start, tup.end, color=color, **kwargs)



import numpy as np
import pandas as pd

from genopandas.core.frame import GenomicDataFrame
from genopandas.core.matrix import AnnotatedMatrix, GenomicMatrix


class CnvValueMatrix(GenomicMatrix):
    """CnvMatrix containing (segmented) logratio values (positions-by-samples).
    """

    def as_segments(self, drop_columns=True):
        """Returns matrix as segments (consecutive stetches with same value).

        Assumes that values have already been segmented, i.e. that bins
        in the same segment have been assigned same numeric value.

        Parameters
        ----------
        drop_columns : bool
            Whether to drop chromosome, start, end and sample columns
            after setting the index.

        Returns
        -------
        GenomicDataFrame
            GenomicDataFrame describing genomic segments. Indexed by
            chromosome, start, end and sample.

            Note that the sample index is included to avoid duplicate index
            errors when reindexing in cases where samples have identical
            segments.
        """

        values = self._values.sort_index()

        # Get segments per sample.
        segment_data = pd.concat(
            (self._segments_for_sample(sample_values)
             for _, sample_values in values.items()),
            axis=0, ignore_index=True)  # yapf: disable

        # Set index. Note that we add sample here, to avoid running into
        # duplicate index errors when re-indexing later down the road.
        # This is a bit of a hack, but GenomicDataFrames shouldn't suffer
        # from having any extra index entries.
        segment_data = segment_data.set_index(
            ['chromosome', 'start', 'end', 'sample'], drop=drop_columns)

        segments = GenomicDataFrame(segment_data)
        segments = segments.gloc[self.gloc.chromosomes]

        return segments

    @staticmethod
    def _segments_for_sample(sample_values):
        # Calculate segment ids (distinguished by diff values).
        segment_ids = np.cumsum(_padded_diff(sample_values) != 0)

        # Get sample and position columns.
        sample = sample_values.name
        chrom_col, start_col, end_col = sample_values.index.names

        # Group and determine positions + values.
        grouped = sample_values.reset_index().groupby(
            by=[chrom_col, segment_ids])

        segments = grouped.agg({
            chrom_col: 'first',
            start_col: 'min',
            end_col: 'max',
            sample: ['first', 'size']
        })

        # Flatten column levels and rename.
        segments.columns = ['_'.join(s) for s in segments.columns]
        segments = segments.rename(columns={
            chrom_col + '_first': 'chromosome',
            start_col + '_min': 'start',
            end_col + '_max': 'end',
            sample + '_first': 'value',
            sample + '_size': 'size'
        })

        # Add sample name and reorder columns.
        segments = segments.reindex(
            columns=['chromosome', 'start', 'end', 'value', 'size'])
        segments['sample'] = sample

        return segments.reset_index(drop=True)

    def to_igv(self, file_path):
        """Saves data for viewing in IGV."""

        igv_data = self._values.reset_index()

        # Rename index columns.
        igv_columns = ['Chromosome', 'Start', 'End']
        column_map = dict(zip(self._values.index.names, igv_columns))

        igv_data = igv_data.rename(columns=column_map)

        # Add 'Feature' column.
        feature_names = ['P{}'.format(i + 1) for i in range(igv_data.shape[0])]
        igv_data.insert(4, 'Feature', feature_names)

        # Write file.
        with open(file_path, 'w') as file_:
            print('#type=COPY_NUMBER', file=file_)
            igv_data.to_csv(file_, sep='\t', index=False, header=True)


class CnvCallMatrix(AnnotatedMatrix):
    """Cnv matrix containing CNV calls (genes-by-samples)."""

    def mask_with_controls(self, column, mask_value=0.0):
        """Masks calls present in control samples.

        Calls are retained if (a) no call is present in the matched control
        sample, (b) if the sample call is more extreme than the control sample
        or (c) the sample and control have calls with different signs
        (loss/gain).

        Matched control samples should be indicated by the given column
        in the sample_data annotation.
        """

        control_samples = self._sample_data[column].dropna()

        new_values = self._values.copy()
        for sample, ctrl in dict(control_samples).items():
            mask = self._call_mask(self._values[ctrl], self._values[sample])
            new_values.loc[~mask, sample] = mask_value

        return self._constructor(new_values)

    @staticmethod
    def _call_mask(ctrl_values, sample_values):
        """Returns mask in which entries are True where ctrl and sample
           have different signs or the sample has a more extreme value.
        """

        ctrl_sign = np.sign(ctrl_values)
        sample_sign = np.sign(sample_values)

        diff_sign = (ctrl_sign - sample_sign).abs() > 1e-8
        higher_val = sample_values.abs() > ctrl_values.abs()

        return diff_sign | (~diff_sign & higher_val)


def _padded_diff(values, pad_value=0):
    """Same as np.diff, with leading 0 to keep same length as input."""
    diff = np.diff(values)
    return np.pad(
        diff, pad_width=(1, 0), mode='constant', constant_values=pad_value)



"""Module containing additional functions for drawing clustermaps."""

from collections import OrderedDict
from itertools import cycle, islice

import pandas as pd

from ._constants import NUMERICAL_CMAPS, CATEGORICAL_COLORS


def color_annotation(df, colors=None, bg_color='#ffffff'):
    """Converts a data frame of annotations to colors."""

    # TODO: Fix boolean/nan case.

    if colors is None:
        colors = generate_colors(df)

    colored_cols = OrderedDict()
    color_maps = OrderedDict()

    for (col_name, values), col_colors in zip(df.items(), colors):
        colored, color_map = _color_column(values, col_colors, bg_color)
        colored_cols[col_name] = colored
        color_maps[col_name] = color_map

    return pd.DataFrame(colored_cols), color_maps


def _color_column(series, colors, bg_color):
    """Converts an annotation column to colors."""

    if series.dtype == bool:
        # Boolean case.
        return _color_bool(series, colors, bg_color)

#     elif series.dtype.kind in 'biufc':
#         # Numeric case.
#         return _color_numeric(series, colors, bg_color)
    elif series.dtype.kind == 'O':
        # Strings and categorical case.
        return _color_categorical(series, colors, bg_color)
    else:
        raise ValueError('Unsupported dtype: {}'.format(series.dtype))


def _color_bool(series, color, bg_color):
    """Converts a boolean annotation column to colors."""

    color_map = OrderedDict(zip([True, False], [color, bg_color]))
    mapped = series.map(color_map).fillna(bg_color)

    return mapped, color_map


def _color_categorical(series, colors, bg_color):
    """Converts a categorical annotation column to colors."""

    if series.dtype.name == 'category':
        str_values = series.cat.categories
    else:
        str_values = set(series.dropna())

    if isinstance(colors, dict):
        color_map = {k: v for k, v in colors.items() if k in str_values}
    else:
        color_map = OrderedDict(zip(str_values, colors))

    colored = series.map(color_map).fillna(bg_color)

    return colored, color_map


# def _color_numeric(series, color, bg_color, n=200):
#     """Converts a numeric annotation column to colors."""

#     cmap = LinearSegmentedColormap.from_list(
#         name='custom_cmap', colors=[bg_color, color], N=n)
#     norm = Normalize(vmin=series.min(), vmax=series.max())
#     mappable = ScalarMappable(norm=norm, cmap=cmap)

#     rgba_colors = mappable.to_rgba(series.values)
#     hex_colors = (rgb2hex(rgba) for rgba in rgba_colors)

#     mapped = pd.Series(hex_colors, index=series.index, name=series.name)

#     return mapped, color


def _rgb_to_hex(rgb, normalized=True):
    if normalized:
        rgb = tuple(map(lambda x: int(x * 255), rgb))
    return '#%02x%02x%02x' % rgb


def generate_colors(annotation,
                    categorical_colors=CATEGORICAL_COLORS,
                    numerical_colors=NUMERICAL_CMAPS):
    """Generate colors for given annotation."""

    # Convert to iterators and cycle.
    categorical_colors = cycle(categorical_colors)
    numerical_colors = cycle(numerical_colors)

    # Assign colors.
    colors = []
    for _, values in annotation.items():
        if values.dtype == bool:
            colors.append(next(categorical_colors))
        elif values.dtype.kind == 'O':
            n_levels = values.nunique()
            colors.append(list(islice(categorical_colors, n_levels)))
        else:
            raise NotImplementedError()

    return colors


def draw_legends(cm, color_maps, margin=1, y=0.6, **kwargs):
    """Draws annotation legends on the given cluster map."""

    fig = cm.fig
    fig.draw(fig.canvas.get_renderer())

    # Draw 'dummy' legends.
    for name, color_map in color_maps.items():
        _draw_legend(fig, color_map, title=name, loc='upper right', **kwargs)

    # Draw to get proper extents.
    fig.draw(fig.canvas.get_renderer())

    # Resize figure to fit legend and labels.
    labels = cm.ax_heatmap.get_yticklabels()

    if cm.ax_col_colors is not None:
        labels += cm.ax_col_colors.get_yticklabels()

    x_inv = _resize_to_fit_legends(
        fig, fig.legends, labels=labels, margin=margin)

    # Get heights and clear legends.
    legend_heights = [l.get_window_extent().height for l in fig.legends]
    fig.legends = []

    # Re-draw final legends.
    inv = fig.transFigure.inverted()
    y_scale = inv.transform([[0, 1], [0, 0]])[0, 1]

    for (name, color_map), height in zip(color_maps.items(), legend_heights):
        _draw_legend(fig, color_map, title=name, loc=(x_inv, y), **kwargs)
        y = y - (height * y_scale)


def _draw_legend(fig, color_map, **kwargs):
    from matplotlib.patches import Patch

    patches = [Patch(color=color, label=label)
               for label, color in color_map.items()]  # yapf: disable

    return fig.legend(patches, color_map.keys(), **kwargs)


def _resize_to_fit_legends(fig, legends, labels, margin=0.5):
    legend_width = max(l.get_window_extent().width for l in legends) / fig.dpi

    if labels:
        label_width = max(l.get_window_extent().width
                          for l in labels) / fig.dpi
    else:
        label_width = 0

    orig_width = fig.get_figwidth()
    required_width = orig_width + label_width + legend_width + margin

    fig.set_figwidth(required_width)
    fig.draw(fig.canvas.get_renderer())

    fig.subplots_adjust(right=orig_width / required_width)
    fig.draw(fig.canvas.get_renderer())

    x_offset = (orig_width + label_width + (margin / 2)) / required_width

    return x_offset



"""Dataframe-related functions/classes."""

from collections import OrderedDict
import math
import re

from natsort import natsorted
import numpy as np
import pandas as pd

from .tree import GenomicIntervalTree


class GenomicDataFrame(pd.DataFrame):
    """DataFrame with fast indexing by genomic position.

    Requires columns 'chromosome', 'start' and 'end' to be present in the
    DataFrame, as these columns are used for indexing.

    Examples
    --------

    Constructing from scratch:

    >>> df = pd.DataFrame.from_records(
    ...    [('1', 10, 20, 'a'), ('2', 10, 20, 'b'), ('2', 30, 40, 'c')],
    ...    columns=['chromosome', 'start', 'end'])
    >>> df.set_index(['chromosome', 'start', 'end'])
    >>> GenomicDataFrame(df)

    Reading from a GTF file:

    >>> GenomicDataFrame.from_gtf('/path/to/reference.gtf')

    Querying by a genomic range:

    >>> genomic_df.gloc['2'][30:50]

    """

    _internal_names = pd.DataFrame._internal_names + ['_gloc']
    _internal_names_set = set(_internal_names)

    _metadata = ['_chrom_lengths']

    def __init__(self, *args, chrom_lengths=None, **kwargs):
        super().__init__(*args, **kwargs)

        self._gloc = None
        self._chrom_lengths = chrom_lengths

    @property
    def _constructor(self):
        return GenomicDataFrame

    @classmethod
    def from_df(cls, df, **kwargs):
        """Constructs instance from dataframe containing
           ranged/positioned data."""

        if df.index.nlevels == 3:
            return cls(df, **kwargs)
        elif df.index.nlevels == 2:
            return cls.from_position_df(df, **kwargs)
        else:
            raise ValueError('DataFrame should have either two index levels '
                             '(for positioned data) or three index levels '
                             '(for ranged data)')

    @classmethod
    def from_position_df(cls, df, width=1, **kwargs):
        """Constructs instance from positioned dataframe."""

        positions = df.index.get_level_values(1)

        starts = positions - (width // 2)
        ends = positions + math.ceil(width / 2)

        new_df = df.copy()
        new_df.index = pd.MultiIndex.from_arrays(
            [df.index.get_level_values(0), starts, ends],
            names=['chromosome', 'start', 'end'])

        return cls(new_df, **kwargs)

    @classmethod
    def from_records(cls,
                     records,
                     index_col,
                     columns=None,
                     drop_index_col=True,
                     chrom_lengths=None,
                     **kwargs):
        """Creates a GenomicDataFrame from a structured or record ndarray."""

        if not 2 <= len(index_col) <= 3:
            raise ValueError('index_col should contain 2 entries'
                             ' (for positioned data or 3 entries'
                             ' (for ranged data)')

        df = super().from_records(records, columns=columns, **kwargs)

        # Convert chromosome to str.
        df[index_col[0]] = df[index_col[0]].astype(str)

        df = df.set_index(index_col, drop=drop_index_col)

        return cls.from_df(df, chrom_lengths=chrom_lengths)

    @classmethod
    def from_csv(cls,
                 file_path,
                 index_col,
                 drop_index_col=True,
                 chrom_lengths=None,
                 **kwargs):
        """Creates a GenomicDataFrame from a csv file using ``pandas.read_csv``.

        Parameters
        ----------
        file_path : str
            Path to file.
        index_col : List[str]
            Columns to use for index. Columns should be indicated by their name.
            Should contain two entries for positioned data, three
            entries for ranged data. If not given, the first three columns of
            the DataFrame are used by default.
        drop_index_col : bool
            Whether to drop the index columns in the DataFrame (True, default)
            or to drop them from the dataframe (False).
        chrom_lengths : Dict[str, int]
            Chromosome lengths.
        **kwargs
            Any extra kwargs are passed to ``pandas.read_csv``.

        Returns
        -------
        GenomicDataFrame
            DataFrame containing the file contents.

        """

        if not 2 <= len(index_col) <= 3:
            raise ValueError('index_col should contain 2 entries'
                             ' (for positioned data or 3 entries'
                             ' (for ranged data)')

        df = pd.read_csv(
            file_path, index_col=None, dtype={index_col[0]: str}, **kwargs)
        df = df.set_index(index_col, drop=drop_index_col)

        return cls.from_df(df, chrom_lengths=chrom_lengths)

    @classmethod
    def from_gtf(cls, gtf_path, filter_=None):
        """Creates a GenomicDataFrame from a GTF file."""

        if (gtf_path.suffixes[-1] == '.gz' and
            gtf_path.with_suffix('.gz.tbi').exists()):
            gdf = cls._from_gtf_pysam(gtf_path, filter_=filter_)
        else:
            gdf = cls._from_gtf_pandas(gtf_path, filter_=filter_)

        return gdf

    @classmethod
    def _from_gtf_pandas(cls, gtf_path, filter_=None):
        """Reads GTF directly using pandas."""

        # Read gtf data.
        data = pd.read_csv(
            gtf_path, sep='\t', header=None, index_col=None,
            names=('contig', 'source', 'feature', 'start', 'end', 'score',
                   'strand', 'frame', 'attributes'),
            dtype={'contig': str})

        # Expand attributes.
        regex = re.compile(r'(?P<key>\w+) "(?P<value>\w+)"')

        rows = (dict(regex.findall(el)) for el in data['attributes'])
        attr_data = pd.DataFrame.from_records(rows)

        data = pd.concat([data.drop('attributes', axis=1), attr_data], axis=1)
        data = data.set_index(['contig', 'start', 'end'], drop=False)

        # Filter rows if filter_ is given.
        if filter_ is not None:
            data = data.query(filter_)

        return cls(data)

    @classmethod
    def _from_gtf_pysam(cls, gtf_path, filter_=None):
        """Reads GTF more efficiently using pysam."""

        try:
            import pysam
        except ImportError:
            raise ImportError('Pysam needs to be installed for '
                              'reading GTF files')

        # Parse records into rows.
        gtf_file = pysam.TabixFile(str(gtf_path), parser=pysam.asGTF())
        records = (rec for rec in gtf_file.fetch())

        # Filter records if needed.
        if filter_ is not None:
            records = (rec for rec in records if filter_(rec))

        # Build dataframe.
        def _record_to_row(record):
            row = {
                'contig': record.contig,
                'source': record.source,
                'feature': record.feature,
                'start': record.start,
                'end': record.end,
                'score': record.score,
                'strand': record.strand,
                'frame': record.frame
            }
            row.update(dict(record))
            return row

        gdf = cls.from_records(
            (_record_to_row(rec) for rec in records),
            index_col=['contig', 'start', 'end'],
            drop_index_col=False)

        # Reorder columns to correspond with GTF format.
        columns = ('contig', 'source', 'feature', 'start', 'end', 'score',
                   'strand', 'frame')
        gdf = _reorder_columns(gdf, order=columns)

        return gdf

    @property
    def gloc(self):
        """Genomic indexer for querying the dataframe."""

        if self._gloc is None:
            self._gloc = GenomicIndexer(self)

        return self._gloc

    @property
    def chromosome_lengths(self):
        """Chromosome lengths."""

        if self._chrom_lengths is None:
            chrom_lengths = self._calculate_chrom_lengths()
            self._chrom_lengths = self._order_chrom_lengths(chrom_lengths)

        return self._chrom_lengths

    @chromosome_lengths.setter
    def chromosome_lengths(self, value):
        if not isinstance(value, OrderedDict):
            value = self._order_chrom_lengths(value)
        self._chrom_lengths = value

    def _calculate_chrom_lengths(self):
        chromosomes = self.index.get_level_values(0)
        ends = self.index.get_level_values(2) - 1

        lengths = pd.Series(ends).groupby(chromosomes).max()
        return dict(zip(lengths.index, lengths.values))

    @staticmethod
    def _order_chrom_lengths(chrom_lengths):
        if not isinstance(chrom_lengths, OrderedDict):
            order = natsorted(chrom_lengths.keys())
            values = (chrom_lengths[k] for k in order)
            chrom_lengths = OrderedDict(zip(order, values))
        return chrom_lengths

    @property
    def chromosome_offsets(self):
        """Chromosome offsets (used when plotting chromosomes linearly)."""

        # Record offsets in ordered dict.
        sorted_lengths = list(self.chromosome_lengths.values())

        cumsums = np.concatenate([[0], np.cumsum(sorted_lengths)])
        offsets = OrderedDict(zip(self.chromosome_lengths.keys(),
                                  cumsums[:-1]))  # yapf: disable

        # Add special marker for end.
        offsets['_END_'] = cumsums[-1]

        return offsets

    def reset_index(self, level=None, drop=False, col_level=0, col_fill=''):
        """Mirrors pd.DataFrame.reset_index, but returns a standard DataFrame
        instead of a GenomicDataFrame, as the (genomic) index is being dropped.
        """

        reset_values = super().reset_index(
            level=level, drop=drop, col_level=col_level, col_fill=col_fill)

        return pd.DataFrame(reset_values)

    def rename_chromosomes(self, mapping):
        """Returns copy with renamed chromosomes."""

        if callable(mapping):
            map_func = mapping
        else:
            map_func = lambda s: mapping.get(s, s)

        # Map chromosomes.
        chrom_col, start_col, end_col = self.index.names

        new_values = self.reset_index()
        new_values[chrom_col] = new_values[chrom_col].map(map_func)

        new_values.set_index([chrom_col, start_col, end_col], inplace=True)

        return self.__class__(new_values)


class GenomicIndexer(object):
    """Base GenomicIndexer class used to index GenomicDataFrames."""

    def __init__(self, gdf):
        self._gdf = gdf
        self._trees = None

    def __getitem__(self, item):
        """Accessor used to query the dataframe by position.

        If a list of chromosomes is given, the dataframe is subset to the
        given chromosomes. Note that chromosomes are also re-ordered to
        adhere to the given order. If a single chromosome is given, a
        GenomicSlice is returned. This slice object can be sliced to query
        a specific genomic range.
        """

        if isinstance(item, list):
            # Add extra index level to enforce uniqueness.
            augmented = self._gdf.set_index(
                self._gdf.groupby(level=0).cumcount(), append=True)

            # Reindex and drop added level.
            subset = augmented.reindex(index=item, level=0)
            subset.index = subset.index.droplevel(subset.index.nlevels - 1)

            # Remove any duplicates.
            items = list(OrderedDict.fromkeys(item))

            # Subset lengths.
            prev_lengths = subset.chromosome_lengths
            subset.chromosome_lengths = OrderedDict(
                (k, prev_lengths[k]) for k in items)  # yapf: disable

            return subset

        return GenomicSlice(self, chromosome=item)

    @property
    def gdf(self):
        """The indexed DataFrame."""
        return self._gdf

    @property
    def chromosome(self):
        """Chromosome values."""
        return self._gdf.index.get_level_values(0)

    @property
    def chromosomes(self):
        """Available chromosomes."""
        return list(self.chromosome_lengths.keys())

    @property
    def chromosome_lengths(self):
        """Chromosome lengths."""
        return self._gdf.chromosome_lengths

    @property
    def chromosome_offsets(self):
        """Chromosome offsets."""
        return self._gdf.chromosome_offsets

    @property
    def start(self):
        """Start positions."""
        return self._gdf.index.get_level_values(1)

    @property
    def start_offset(self):
        """Start positions, offset by chromosome lengths."""
        return self._offset_positions(self.start)

    @property
    def end(self):
        """End positions."""
        return self._gdf.index.get_level_values(2)

    @property
    def end_offset(self):
        """End positions, offset by chromosome lengths."""
        return self._offset_positions(self.end)

    @property
    def position(self):
        """Mid positions (between start/end).

        Should corrrespond with original positions for
        expanded positioned data.
        """
        return (self.start + self.end) // 2

    @property
    def position_offset(self):
        """Mid positions (see position), offset by chromosome lengths."""
        return self._offset_positions(self.position)

    def _offset_positions(self, positions):
        offsets = pd.Series(self.chromosome_offsets)
        return positions + offsets.loc[self.chromosome].values

    @property
    def trees(self):
        """Trees used for indexing the DataFrame."""

        if self._trees is None:
            self._trees = self._build_trees()

        return self._trees

    def rebuild(self):
        """Rebuilds the genomic interval trees."""
        self._trees = self._build_trees()

    def _build_trees(self):
        tuples = zip(self.chromosome, self.start, self.end,
                     range(self._gdf.shape[0]))
        return GenomicIntervalTree.from_tuples(tuples)

    def search(self,
               chromosome,
               start,
               end,
               strict_left=False,
               strict_right=False):
        """Searches the DataFrame for rows within given range."""

        overlap = self.trees.search(
            chromosome,
            start,
            end,
            strict_left=strict_left,
            strict_right=strict_right)

        indices = [interval[2] for interval in overlap]

        return self._gdf.iloc[indices].sort_index()


class GenomicSlice(object):
    """Supporting class used by the GenomicIndexer for slicing chromosomes."""

    def __init__(self, indexer, chromosome):
        self._indexer = indexer
        self._chromosome = chromosome

    def __getitem__(self, item):
        if isinstance(item, slice):
            subset = self._indexer.search(
                self._chromosome, start=item.start, end=item.stop)

            # Subset lengths.
            subset.chromosome_lengths = OrderedDict(
                [(self._chromosome,
                  subset.chromosome_lengths[self._chromosome])])

            return subset

        return self._indexer.search(self._chromosome, start=item)


def _reorder_columns(df, order):
    """Reorders dataframe columns, sorting any extra columns alphabetically."""
    extra_cols = set(df.columns) - set(order)
    return df[list(order) + sorted(extra_cols)]



from pkg_resources import get_distribution, DistributionNotFound
try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:
    # package is not installed
    pass

from .core import *

__author__ = 'Julian de Ruiter'
__email__ = 'julianderuiter@gmail.com'



import functools

import pandas as pd


class DfWrapper(object):
    """Utility class used to build wrapper classes for pandas DataFrames."""

    def __init__(self, values):
        self._values = values

    @property
    def values(self):
        """Internal DataFrame containing values."""
        return self._values

    @property
    def loc(self):
        """Label-based indexer (similar to pandas .loc)."""
        return DfLocWrapper(self._values.loc, constructor=self._constructor)

    @property
    def iloc(self):
        """Index-based indexer (similar to pandas .iloc)."""
        return DfLocWrapper(self._values.iloc, constructor=self._constructor)

    def _constructor(self, values):
        """Constructor that attempts to build new instance
           from given values."""

        if isinstance(values, pd.DataFrame):
            return self.__class__(values.copy())

        return values

    def __getitem__(self, item):
        values = self._values[item]
        return self._constructor(values)

    def __getattr__(self, name):
        if hasattr(self._values, name):
            attr = getattr(self._values, name)
        else:
            raise AttributeError('{!r} object has no attribute {!r}'.format(
                self.__class__.__name__, name))

        if callable(attr):
            attr = self._wrap_function(attr)
        else:
            attr = self._constructor(attr)

        return attr

    def _wrap_function(self, func):
        """Wrap functions to call _constructor on returned value."""

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            result = func(*args, **kwargs)
            return self._constructor(result)

        return wrapper


class DfLocWrapper(object):
    """Wrapper class that wraps an objects loc/iloc accessor."""

    def __init__(self, loc, constructor=None):
        if constructor is None:
            constructor = lambda x: x

        self._loc = loc
        self._constructor = constructor

    def __getitem__(self, item):
        result = self._loc[item]
        return self._constructor(result)



import numpy as np
import pandas as pd
import pytest

from genopandas.core.frame import GenomicDataFrame
from genopandas.core.matrix import AnnotatedMatrix, GenomicMatrix


@pytest.fixture
def feature_values(random):
    """Example feature matrix values."""

    return pd.DataFrame(
        np.random.randn(4, 4),
        columns=['s1', 's2', 's3', 's4'],
        index=['f1', 'f2', 'f3', 'f4'])


@pytest.fixture
def sample_data():
    """Example sample data."""
    return pd.DataFrame(
        {
            'phenotype': ['sensitive', 'resistant'] * 2
        },
        index=['s1', 's2', 's3', 's4'])


@pytest.fixture
def feature_data():
    """Example feature data."""

    return pd.DataFrame(
        {
            'biotype': ['a', 'a', 'b', 'b']
        }, index=['f1', 'f2', 'f3', 'f4'])


@pytest.fixture
def example_matrix(feature_values, sample_data, feature_data):
    """Example feature matrix."""

    return AnnotatedMatrix(
        feature_values, sample_data=sample_data, feature_data=feature_data)


class TestAnnotatedMatrix(object):
    """Tests for AnnotatedMatrix class."""

    def test_init(self, feature_values, sample_data, feature_data):
        """Tests for basic init."""

        matrix = AnnotatedMatrix(
            feature_values, sample_data=sample_data, feature_data=feature_data)

        assert all(matrix.values == feature_values)
        assert list(matrix.columns) == ['s1', 's2', 's3', 's4']
        assert list(matrix.index) == ['f1', 'f2', 'f3', 'f4']

        assert all(matrix.sample_data == sample_data)
        assert all(matrix.feature_data == feature_data)

    def test_init_without_samples(self, feature_values):
        """Tests for init without sample_data."""

        matrix = AnnotatedMatrix(feature_values)

        assert all(matrix.values == feature_values)
        assert list(matrix.columns) == ['s1', 's2', 's3', 's4']
        assert list(matrix.index) == ['f1', 'f2', 'f3', 'f4']

        dummy = pd.DataFrame({}, index=['s1', 's2', 's3', 's4'])
        assert all(matrix.sample_data == dummy)

    def test_loc(self, example_matrix):
        """Tests subsetting matrix using loc."""

        # Test subsetting with list.
        subset = example_matrix.loc[['f2', 'f3']]
        assert list(subset.index) == ['f2', 'f3']

        # Test single element, should return series.
        subset2 = example_matrix.loc['f2']
        assert isinstance(subset2, pd.Series)

    def test_iloc(self, example_matrix):
        """Tests subsetting matrix using iloc."""

        # Test subsetting with list.
        subset = example_matrix.iloc[[0, 1]]
        assert list(subset.index) == ['f1', 'f2']
        assert list(subset.columns) == ['s1', 's2', 's3', 's4']

        # Test single element, should return series.
        subset2 = example_matrix.iloc[1]
        assert isinstance(subset2, pd.Series)

    def test_get_item(self, example_matrix):
        """Tests subsetting samples using get_item."""

        # Tests subsetting with list.
        subset = example_matrix[['s2', 's3']]
        assert list(subset.columns) == ['s2', 's3']

        # Tests single element.
        subset2 = example_matrix['s2']
        assert isinstance(subset2, pd.Series)

    def test_rename(self, example_matrix):
        """Tests renaming."""

        # Test renaming of samples.
        renamed = example_matrix.rename(columns={'s2': 's5'})
        assert list(renamed.columns) == ['s1', 's5', 's3', 's4']
        assert list(renamed.values.columns) == ['s1', 's5', 's3', 's4']
        assert list(renamed.sample_data.index) == ['s1', 's5', 's3', 's4']

        # Test renaming of features.
        renamed = example_matrix.rename(index={'f2': 'f5'})
        assert list(renamed.index) == ['f1', 'f5', 'f3', 'f4']
        assert list(renamed.values.index) == ['f1', 'f5', 'f3', 'f4']
        assert list(renamed.feature_data.index) == ['f1', 'f5', 'f3', 'f4']

    def test_query(self, example_matrix):
        """Tests subsetting samples with query_samples."""

        subset = example_matrix.query_samples('phenotype == "sensitive"')
        assert list(subset.columns) == ['s1', 's3']

        assert list(subset.values.columns) == ['s1', 's3']
        assert list(subset.sample_data.index) == ['s1', 's3']

    def test_dropna_samples(self, feature_values, feature_data):
        """Tests dropping samples with NAs in sample_data."""

        sample_data = pd.DataFrame(
            {
                'phenotype': ['sensitive', None, None, 'resistant']
            },
            index=['s1', 's2', 's3', 's4'])

        matrix = AnnotatedMatrix(
            feature_values, sample_data=sample_data, feature_data=feature_data)

        matrix = matrix.dropna_samples()

        assert list(matrix.columns) == ['s1', 's4']
        assert list(matrix.sample_data.index) == ['s1', 's4']

    def test_dropna(self, example_matrix):
        """Tests dropping features with NAs in values."""

        example_matrix.values.loc['f2', 's2'] = None

        matrix2 = example_matrix.dropna(axis=0)
        assert list(matrix2.index) == ['f1', 'f3', 'f4']
        assert list(matrix2.columns) == ['s1', 's2', 's3', 's4']

        matrix3 = example_matrix.dropna(axis=1)
        assert list(matrix3.index) == ['f1', 'f2', 'f3', 'f4']
        assert list(matrix3.columns) == ['s1', 's3', 's4']
        assert list(matrix3.sample_data.index) == ['s1', 's3', 's4']

    def test_from_csv(self, sample_data):
        """Tests from_csv."""

        file_path = pytest.helpers.data_path('matrix_features.tsv')

        matrix = AnnotatedMatrix.from_csv(
            file_path, sep='\t', sample_data=sample_data)

        assert list(matrix.columns) == ['s1', 's2', 's3', 's4']
        assert list(matrix.index) == ['f1', 'f2', 'f3', 'f4']

        assert list(matrix.sample_data.index) == ['s1', 's2', 's3', 's4']
        assert list(matrix.sample_data.columns) == ['phenotype']

    def test_from_csv_extra(self, sample_data):
        """Tests from_csv with an extra column."""

        file_path = pytest.helpers.data_path('matrix_features_extra.tsv')

        matrix = AnnotatedMatrix.from_csv(
            file_path, sep='\t', sample_data=sample_data, drop_cols=['extra'])

        assert list(matrix.columns) == ['s1', 's2', 's3', 's4']
        assert list(matrix.index) == ['f1', 'f2', 'f3', 'f4']


@pytest.fixture
def genomic_values(random):
    """Example feature matrix values."""

    index = pd.MultiIndex.from_tuples(
        [('1', 20, 30), ('1', 30, 40), ('2', 10, 25), ('2', 50, 60)],
        names=['chromosome', 'start', 'end'])

    return pd.DataFrame(
        np.random.randn(4, 4), columns=['s1', 's2', 's3', 's4'], index=index)


@pytest.fixture
def genomic_matrix(genomic_values, sample_data):
    """Example feature matrix."""

    return GenomicMatrix(genomic_values, sample_data=sample_data)


@pytest.fixture
def logratio_matrix():
    """Example dataset containing logratios."""

    file_path = pytest.helpers.data_path('logratios.tsv')

    return GenomicMatrix.from_csv_condensed(
        file_path, sep='\t', is_one_based=True, is_inclusive=False)


@pytest.fixture
def genes():
    """Example genes."""

    file_path = pytest.helpers.data_path('genes.tsv')

    return GenomicDataFrame.from_csv(
        file_path, sep='\t', index_col=['chromosome', 'start', 'end'])


class TestGenomicMatrix(object):
    """Tests for GenomicMatrix base class."""

    def test_init(self, genomic_values, sample_data):
        """Tests basic init."""
        matrix = GenomicMatrix(genomic_values, sample_data)
        assert all(matrix.values == genomic_values)
        assert all(matrix.sample_data == sample_data)

    def test_gloc_subset(self, genomic_matrix):
        """Tests subsetting chromosomes using gloc."""

        subset = genomic_matrix.gloc[['1']]
        assert list(genomic_matrix.values.gloc.chromosomes) == ['1', '2']
        assert list(subset.values.gloc.chromosomes) == ['1']

    def test_gloc_reorder(self, genomic_matrix):
        """Tests reordering chromosomes using gloc."""

        reordered = genomic_matrix.gloc[['2', '1']]

        assert list(genomic_matrix.values.gloc.chromosomes) == ['1', '2']
        assert list(reordered.values.gloc.chromosomes) == ['2', '1']

    def test_gloc_search(self, genomic_matrix):
        """Tests search using gloc."""

        subset = genomic_matrix.gloc.search('1', 10, 30)
        assert isinstance(subset, GenomicMatrix)
        assert subset.shape[0] == 1

    def test_gloc_slice(self, genomic_matrix):
        """Tests slicing using gloc."""

        subset = genomic_matrix.gloc['1'][10:30]
        assert subset.shape[0] == 1

        assert subset == genomic_matrix.gloc.search('1', 10, 30)

    def test_gloc_attributes(self, genomic_matrix):
        """Tests (wrapped) accessor attributes on gloc."""

        values = genomic_matrix.values

        assert all(genomic_matrix.gloc.chromosome == values.gloc.chromosome)
        assert all(genomic_matrix.gloc.start == values.gloc.start)
        assert all(genomic_matrix.gloc.end == values.gloc.end)

        assert genomic_matrix.gloc.chromosome_lengths == \
                   values.gloc.chromosome_lengths
        assert genomic_matrix.gloc.chromosome_offsets == \
                   values.gloc.chromosome_offsets

    def test_from_csv(self):
        """Tests reading using from_csv."""

        file_path = pytest.helpers.data_path('matrix_ranged.tsv')

        matrix = GenomicMatrix.from_csv(
            file_path, sep='\t', index_col=['chromosome', 'start', 'end'])

        assert list(matrix.columns) == ['s1', 's2', 's3', 's4']
        assert list(matrix.gloc.chromosome) == ['1', '1', '2', '2']
        assert list(matrix.gloc.start) == [20, 30, 10, 50]
        assert list(matrix.gloc.end) == [30, 40, 25, 60]

    def test_from_csv_condensed(self):
        """Tests reading using from_csv_condensed."""

        file_path = pytest.helpers.data_path('matrix_ranged_condensed.tsv')

        matrix = GenomicMatrix.from_csv_condensed(file_path, sep='\t')

        assert list(matrix.columns) == ['s1', 's2', 's3', 's4']
        assert list(matrix.gloc.chromosome) == ['1', '1', '2', '2']
        assert list(matrix.gloc.start) == [20, 30, 10, 50]
        assert list(matrix.gloc.end) == [30, 40, 25, 60]

    def test_from_csv_condensed_pos(self):
        """Tests reading using from_csv_condensed with positions."""

        file_path = pytest.helpers.data_path('matrix_ranged_condensed_pos.tsv')

        matrix = GenomicMatrix.from_csv_condensed(
            file_path,
            sep='\t',
            index_regex=r'(?P<chromosome>\w+):(?P<position>\d+)')

        assert list(matrix.columns) == ['s1', 's2', 's3', 's4']
        assert list(matrix.gloc.chromosome) == ['1', '1', '2', '2']
        assert list(matrix.gloc.start) == [20, 30, 10, 50]
        assert list(matrix.gloc.end) == [21, 31, 11, 51]
        assert list(matrix.gloc.position) == [20, 30, 10, 50]

    def test_expand(self, logratio_matrix):
        """Test expansion using expand."""

        assert logratio_matrix.shape[0] == 9

        # Check shape.
        expanded = logratio_matrix.expand()
        assert expanded.shape[0] == 10

        # Check imputed row.
        assert expanded.index[2] == ('1', 3100000, 3150000)
        assert all(expanded.iloc[2].isnull())

        # Check not everything is null (misaligned index).
        assert not expanded.values.isnull().all().all()

    def test_impute(self, logratio_matrix):
        """Tests imputation."""

        assert logratio_matrix.shape[0] == 9

        # Check shape.
        imputed = logratio_matrix.impute()
        assert imputed.shape[0] == 10

        # Check imputed row.
        assert imputed.index[2] == ('1', 3100000, 3150000)
        assert not any(imputed.iloc[2].isnull())

        # Check not everything is null (misaligned index).
        assert not imputed.values.isnull().all().all()

    def test_resample(self, logratio_matrix):
        """Tests resampling."""

        # Check shape.
        resampled = logratio_matrix.resample(bin_size=100000)
        assert resampled.shape[0] == 5

        # Check indices.
        assert list(resampled.gloc.start) == [
            3000000, 3100000, 3200000, 3300000, 3400000
        ]
        assert list(resampled.gloc.end) == [
            3100000, 3200000, 3300000, 3400000, 3500000
        ]

        # Check values.
        assert not resampled.isnull().all().all()

    def test_annotate(self, logratio_matrix, genes):
        """Tests annotation of logratio matrix."""

        annotated = logratio_matrix.annotate(genes, feature_id='name')

        assert list(annotated.index) == ['gene a']
        assert list(annotated.columns) == list(logratio_matrix.columns)

        expected = logratio_matrix.iloc[:2].mean(axis=0).values
        assert all((annotated.values.iloc[0].values - expected) < 1e-6)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

REQUIREMENTS = ['intervaltree', 'pandas', 'numpy', 'natsort', 'toolz']

setup(
    name='genopandas',
    description='',
    author='Julian de Ruiter',
    author_email='julianderuiter@gmail.com',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    include_package_data=True,
    install_requires=REQUIREMENTS,
    setup_requires=['setuptools_scm'],
    use_scm_version={'root': '.', 'relative_to': __file__},
    license='MIT license',
    zip_safe=False,
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.5',
    ])



"""Tests for pandas-related classes/functions."""

import numpy as np
import pandas as pd
from pandas.api.types import is_numeric_dtype
import pytest

from genopandas.core.frame import GenomicDataFrame

# pylint: disable=redefined-outer-name,no-self-use


@pytest.fixture()
def ranged_df():
    """Simple example of a ranged dataframe."""

    index = pd.MultiIndex.from_tuples(
        [('1', 20, 30), ('1', 30, 40), ('2', 10, 25), ('2', 50, 60)],
        names=['chromosome', 'start', 'end'])

    df = pd.DataFrame(
        np.random.randn(4, 4),
        columns=['s{}'.format(i + 1) for i in range(4)],
        index=index)

    return df


@pytest.fixture()
def positioned_df():
    """Simple example of a ranged dataframe."""

    index = pd.MultiIndex.from_tuples(
        [('1', 20), ('1', 30), ('2', 10), ('2', 50)],
        names=['chromosome', 'position'])

    df = pd.DataFrame(
        np.random.randn(4, 4),
        columns=['s{}'.format(i + 1) for i in range(4)],
        index=index)

    return df


@pytest.fixture()
def ranged_gdf(ranged_df):
    """Simple example genomic dataframe."""
    return GenomicDataFrame(ranged_df)


class TestGenomicDataFrame(object):
    """Tests for GenomicDataFrame class."""

    def test_init(self, ranged_df):
        """Tests init with example data."""

        gdf = GenomicDataFrame(ranged_df)

        # Check shape of frame.
        assert len(gdf) == 4
        assert list(gdf.columns) == ['s1', 's2', 's3', 's4']

        # Check index access via named properties.
        assert list(gdf.gloc.chromosome) == ['1', '1', '2', '2']
        assert list(gdf.gloc.start) == [20, 30, 10, 50]
        assert list(gdf.gloc.end) == [30, 40, 25, 60]
        assert list(gdf.gloc.position) == [25, 35, 17, 55]

        # Check available chromosomes.
        assert gdf.gloc.chromosomes == ['1', '2']

    def test_gloc_subset(self, ranged_gdf):
        """Tests subsetting chromosomes using gloc."""

        subset = ranged_gdf.gloc[['2']]

        assert len(subset) == 2
        assert list(subset.gloc.chromosome) == ['2', '2']
        assert subset.gloc.chromosomes == ['2']

    def test_gloc_reorder(self, ranged_gdf):
        """Tests reordering chromosomes using gloc."""

        subset = ranged_gdf.gloc[['2', '1']]

        assert len(subset) == 4
        assert list(subset.gloc.chromosome) == ['2', '2', '1', '1']
        assert subset.gloc.chromosomes == ['2', '1']

    def test_gloc_slice(self, ranged_gdf):
        """Tests slicing of dataframe using gloc."""

        subset = ranged_gdf.gloc['1'][10:30]
        assert len(subset) == 1

    def test_gloc_search(self, ranged_gdf):
        """Test searching of dataframe using gloc."""

        # Test same example as slice.
        subset = ranged_gdf.gloc.search('1', start=10, end=30)
        assert len(subset) == 1

        # Test strict search with example within bounds...
        subset = ranged_gdf.gloc.search(
            '1', start=10, end=30, strict_right=True)
        assert len(subset) == 1

        # ...and extending beyond bounds.
        subset = ranged_gdf.gloc.search(
            '1', start=10, end=29, strict_right=True)
        assert len(subset) == 0

    def test_gloc_lengths(self, ranged_gdf):
        """Tests computation of chromosome lengths."""

        expected = {'1': 39, '2': 59}
        assert ranged_gdf.gloc.chromosome_lengths == expected

    def test_gloc_lengths_predefined(self, ranged_df):
        """Tests use of predefined lengths."""

        # Test example.
        ranged_gdf = GenomicDataFrame(
            ranged_df, chrom_lengths={
                '1': 120,
                '2': 90,
            })

        expected = {'1': 120, '2': 90}
        assert ranged_gdf.gloc.chromosome_lengths == expected

        # Test example with extra (unused) chromosome.
        # Should keep the chromosome in lengths.
        ranged_gdf = GenomicDataFrame(
            ranged_df, chrom_lengths={'1': 120,
                                      '2': 90,
                                      '3': 80})

        expected = {'1': 120, '2': 90, '3': 80}
        assert ranged_gdf.gloc.chromosome_lengths == expected

    def test_gloc_offsets(self, ranged_gdf):
        """Tests computation of offsets."""

        # Check offsets.
        expected = {'1': 0, '2': 39, '_END_': 98}
        assert ranged_gdf.gloc.chromosome_offsets == expected

        # Check offset starts/ends.
        assert list(ranged_gdf.gloc.start_offset) == [20, 30, 49, 89]
        assert list(ranged_gdf.gloc.end_offset) == [30, 40, 64, 99]
        assert list(ranged_gdf.gloc.position_offset) == [25, 35, 56, 94]

    def test_gloc_offsets_predefined(self, ranged_df):
        """Tests use of predefined lengths."""

        # Test example.
        ranged_gdf = GenomicDataFrame(
            ranged_df, chrom_lengths={'1': 120,
                                      '2': 90})

        expected = {'1': 0, '2': 120, '_END_': 210}
        assert ranged_gdf.gloc.chromosome_offsets == expected

        # Test example with extra (unused) chromosome.
        # Should not omit chromosome from offsets.
        ranged_gdf = GenomicDataFrame(
            ranged_df, chrom_lengths={'1': 120,
                                      '2': 90,
                                      '3': 80})

        expected = {'1': 0, '2': 120, '3': 210, '_END_': 290}
        assert ranged_gdf.gloc.chromosome_offsets == expected

    def test_from_csv(self):
        """Tests reading data from tsv."""

        # Read file.
        file_path = pytest.helpers.data_path('frame_ranged.tsv')
        gdf = GenomicDataFrame.from_csv(
            file_path, index_col=['chromosome', 'start', 'end'], sep='\t')

        # Check shape.
        assert len(gdf) == 4
        assert list(gdf.columns) == ['s1', 's2', 's3', 's4']

        # Check some dtypes.
        assert not is_numeric_dtype(gdf.index.get_level_values(0))
        assert is_numeric_dtype(gdf.index.get_level_values(1))
        assert is_numeric_dtype(gdf.index.get_level_values(2))

    def test_from_gtf(self):
        """Tests reading data from gtf."""

        # Read file.
        file_path = pytest.helpers.data_path('reference.gtf.gz')
        gdf = GenomicDataFrame.from_gtf(file_path)

        # Check shape.
        assert len(gdf) > 0
        assert list(gdf.columns[:8]) == [
            'contig', 'source', 'feature', 'start', 'end', 'score', 'strand',
            'frame'
        ]

        # Check some dtypes.
        assert not is_numeric_dtype(gdf['contig'])
        assert is_numeric_dtype(gdf['start'])
        assert is_numeric_dtype(gdf['end'])

    def test_from_gtf_filter(self):
        """Tests reading data from gtf with filter."""

        # Read file.
        file_path = pytest.helpers.data_path('reference.gtf.gz')
        gdf = GenomicDataFrame.from_gtf(
            file_path, filter_=lambda rec: rec.feature == 'gene')

        # Check result.
        assert len(gdf) > 0
        assert set(gdf['feature']) == {'gene'}

    def test_from_records_tuple(self):
        """Tests from_records with tuple input."""

        records = [('1', 20, 30, 'gene', 'GeneA'),
                   ('1', 40, 60, 'gene', 'GeneB')]  # yapf: disable

        # Test with named columns.
        gdf = GenomicDataFrame.from_records(
            records,
            columns=['chromosome', 'start', 'end', 'feature', 'name'],
            index_col=['chromosome', 'start', 'end'])

        assert len(gdf) == 2
        assert list(gdf.columns) == ['feature', 'name']

        # Test un-named columns.
        gdf2 = GenomicDataFrame.from_records(records, index_col=[0, 1, 2])
        assert len(gdf2) == 2
        assert list(gdf2.columns) == [3, 4]

    def test_from_records_tuple_dict(self):
        """Tests from_records with dict input."""

        records = [{'chromosome': '1', 'start': 20, 'end': 30,
                    'feature': 'gene', 'name': 'GeneA'},
                   {'chromosome': '1', 'start': 40, 'end': 60,
                    'feature': 'gene', 'name': 'GeneB'}]  # yapf: disable

        gdf = GenomicDataFrame.from_records(
            records, index_col=['chromosome', 'start', 'end'])

        assert len(gdf) == 2
        assert list(gdf.columns) == ['feature', 'name']

    def test_from_position_df(self, positioned_df):
        """Test from_position_df method."""
        gdf = GenomicDataFrame.from_position_df(positioned_df)

        assert list(gdf.gloc.chromosome) == ['1', '1', '2', '2']
        assert list(gdf.gloc.start) == [20, 30, 10, 50]
        assert list(gdf.gloc.end) == [21, 31, 11, 51]
        assert list(gdf.gloc.position) == [20, 30, 10, 50]

    def test_from_df_positioned(self, positioned_df):
        """Test from_df method with positioned data."""

        gdf = GenomicDataFrame.from_df(positioned_df)

        assert list(gdf.gloc.chromosome) == ['1', '1', '2', '2']
        assert list(gdf.gloc.start) == [20, 30, 10, 50]
        assert list(gdf.gloc.end) == [21, 31, 11, 51]
        assert list(gdf.gloc.position) == [20, 30, 10, 50]



"""Module containing functions that supplant seaborn's plotting functions.

Examples include a ``scatter`` function, which creates a scatterplot whilst
keeping an API similar to the other seaborn plotting functions.
"""

from itertools import cycle

import toolz

from ._constants import CATEGORICAL_COLORS


def scatter_plot(data,
                 x,
                 y,
                 hue=None,
                 hue_order=None,
                 palette=None,
                 color=None,
                 legend=True,
                 legend_kws=None,
                 ax=None,
                 **kwargs):
    """Draws a scatter plot of the given dataset.

    Input data is expected to be in long (tidy) form.

    Parameters
    ----------
    data : pd.DataFrame
        Dataset for plotting. Is expected to be in long (tidy) form.
    x, y, hue : str
        Columns to use for plotting. ``x`` and ``y`` determine what is plotted
        along the x- and y-axes, respectively. If ``hue`` is given, points are
        colored according to the (categorical) values of the respective column.
    hue_order : List[str]
        Order to plot the categorical hue levels in.
    palette : List[str] or Dict[Any, str]
        Colors to use for the different levels of the hue variable. Can either
        be a dictionary mapping values to specific colors, or a list of colors
        to use.
    color : matplotlib color
        Color to use for all elements. Overrides palette if given.
    legend : bool
        Whether to draw a legend for the different hue levels.
        (Only used if hue is given.)
    legend_kws : Dict[str, Any]
        Dictionary of additional keyword arguments to pass to ax.legend
        when drawing the legend.
    ax : AxesSubplot
        Axis to use for drawing.
    kwargs : Dict[str, Any]
        Other keyword arguments are passed through to ``ax.plot`` at draw time.

    Returns
    -------
    AxesSubplot
        Axis on which the scatterplot was drawn.

    """

    from matplotlib import pyplot as plt

    default_plot_kws = {'linestyle': 'None', 'marker': 'o'}
    plot_kws = {**default_plot_kws, **kwargs}

    if ax is None:
        _, ax = plt.subplots()

    if hue is None:
        ax.plot(data[x], data[y], color=color, **plot_kws)
    else:
        if palette is None:
            palette = cycle(CATEGORICAL_COLORS)

        data = data.assign(_color=apply_palette(
            data[hue], palette, color=color, order=hue_order))

        # Group data by label/color.
        groups = {
            label: (color, grp)
            for (label, color), grp in data.groupby([hue, '_color'])
        }

        # Plot groups in hue order.
        hue_order = hue_order or set(groups.keys())

        for label in hue_order:
            color, grp = groups[label]
            ax.plot(grp[x], grp[y], label=label, color=color, **plot_kws)

        if legend:
            # Draw legend.
            default_legend_kws = {
                'frameon': True,
                'title': hue,
                'loc': 'upper right'
            }
            ax.legend(**toolz.merge(default_legend_kws, (legend_kws or {})))

    ax.set_xlabel(x)
    ax.set_ylabel(y)

    return ax


def step_plot(data,
              x,
              y,
              hue=None,
              hue_order=None,
              palette=None,
              color=None,
              legend=True,
              legend_kws=None,
              ax=None,
              **kwargs):

    from matplotlib import pyplot as plt

    default_step_kws = {'where': 'post'}
    step_kws = {**default_step_kws, **kwargs}

    if ax is None:
        _, ax = plt.subplots()

    if hue is None:
        ax.step(data[x], data[y], color=color, **step_kws)
    else:
        if palette is None:
            palette = cycle(CATEGORICAL_COLORS)

        data = data.assign(_color=apply_palette(
            data[hue], palette, color=color, order=hue_order))

        for (label, color), grp in data.groupby([hue, '_color']):
            ax.step(grp[x], grp[y], label=label, color=color, **step_kws)

        if legend:
            default_legend_kws = {
                'frameon': True,
                'title': hue,
                'loc': 'upper right'
            }
            ax.legend(**toolz.merge(default_legend_kws, (legend_kws or {})))

    ax.set_xlabel(x)
    ax.set_ylabel(y)

    return ax


def apply_palette(series, palette, color=None, bg_color='white', order=None):
    """Uses palette to color values in the given categorical series.

    Parameters
    ----------
    series : pd.Series
        Values to color.
    palette : List[str] or Dict[Any, str]
        Colors to use for the different levels of the series. Can either be a
        dictionary mapping values to specific colors or a list of colors.
    color : matplotlib color
        Color to use for all elements. Overrides palette if given.
    bg_color : str
        Background color to use for NA values.
    order : list[str]
        Order to color the categorical hue levels in.

    Returns
    -------
    pd.Series
        Series containing colored values.

    """

    if not isinstance(palette, dict):
        if order is None:
            order = series.unique()

        colors = cycle(palette)
        palette = dict(zip(order, colors))

    if color is not None:
        palette = {k: color for k in palette.keys()}

    return series.map(palette).fillna(bg_color)



"""Module containing various plotting constants shared between modules."""

NUMERICAL_CMAPS = ('viridis', 'magma', 'plasma', 'inferno')

CATEGORICAL_COLORS = ('#1f77b4', '#ff7f0e', '#2ca02c', '#d62728', '#9467bd',
                      '#8c564b', '#e377c2', '#7f7f7f', '#bcbd22', '#17becf')
